#!/usr/bin/env python3
# plot toplev -I... -x, -o ...csv output as bar plot
#
from __future__ import print_function
import os
import re
import argparse
from math import isnan, trunc
from collections import defaultdict
import matplotlib
if os.getenv('DISPLAY') is None:
    matplotlib.use('Agg')
import matplotlib.pyplot as plt
import gen_level
import tldata

def parse_args():
    p = argparse.ArgumentParser(usage='plot toplev -I...  -x, output as bar plot')
    p.add_argument('file', help='CSV file to plot')
    p.add_argument('--output', '-o', help='Save figure to file (.pdf/.png/etc). Otherwise show.',
                      nargs='?')
    p.add_argument('--verbose', '-v', help='Plot all data values even if below threshold', action='store_true')
    p.add_argument('--xkcd', help='Enable XKCD mode (with new matplotlib). Please install Humor Sans.', action='store_true')
    p.add_argument('--title', help='Set title of plot', nargs='?')
    p.add_argument('--quiet', help='Be quiet', action='store_true')
    p.add_argument('--cpu', help='CPU to plot (by default first)')  # XXX
    return p.parse_args()

args = parse_args()

try:
    import brewer2mpl
except ImportError:
    if not args.quiet:
        print("pip install brewer2mpl for better colors")

if args.xkcd:
    plt.xkcd()

data = tldata.TLData(args.file, args.verbose)
data.update()

levels = data.levels
timestamps = data.times
ratios = defaultdict(list)
# XXX plot multiple cpus instead
cpu = None
if args.cpu:
    cpu = args.cpu
elif 'CLKS' in data.headers and len(data.vals) > 0:
    # pick CPU with highest utilization. XXX look at all time series
    util = sorted([(data.vals[0][x], x[1]) for x in data.vals[0].keys() if x[0] == 'CLKS'],
                  reverse=True)
    cpu = util[0][1]
elif len(data.cpus) > 0:
    cpu = sorted(sorted(data.cpus), key=len, reverse=True)[0]

def cpumatch(x, cpu, base):
    return x.startswith(cpu) or x == base

if cpu:
    base = None
    m = re.match(r'C\d+', cpu)
    if m:
        base = m.group(0)
    aliases = [x for x in data.cpus if cpumatch(x, cpu, base)]
    print("plotting cpus:", " ".join(sorted(aliases)))
else:
    aliases = []
if len(aliases) == 0:
    aliases = [None]

for h in data.headers:
    def findval(d):
        for c in aliases:
            if (h, c) in d:
                return d[(h, c)]
        return float('nan')

    ratios[h] = list(map(findval, data.vals))

def valid_row(r):
    s = sum(r)
    #if sum([0 if isnan(x) else 1 for x in r]) < len(r)/80.:
    #    return False
    return s != 0.0 and not isnan(s)

def get_colors(non_null):
    if 'brewer2mpl' in globals():
        num_color = max(min(len(non_null), 11), 3)
        all_colors = brewer2mpl.get_map('Spectral', 'Diverging', num_color).hex_colors
    else:
        all_colors = None
    return all_colors

def set_title(ax, t):
    try:
        ax.set_title(t, {'fontsize': 6}, loc='right')
    except AttributeError:
        ax.set_title(t)

def suffix(x):
    dot = x.rfind('.')
    if dot >= 0:
        return x[dot + 1:]
    return x

n = 0
numplots = len(levels.keys())
ax = None
yset = False
max_legend = 0
xaxis = None
legend_bbox = (0., 0., -0.07, -0.03)
legend_loc = 2

for l in tldata.level_order(data):
    non_null = [x for x in levels[l] if valid_row(ratios[x])]
    if not non_null:
        n += 1
        continue
    all_colors = get_colors(non_null)
    ax = plt.subplot2grid((numplots, 1), (n, 0), sharex=xaxis)
    plt.tight_layout()
    set_title(ax, l)
    r = [[y if not isnan(y) else 0.0 for y in ratios[x]] for x in non_null]

    if gen_level.is_metric(non_null[0]):
        for j, name in zip(r, non_null):
            stack = ax.plot(timestamps, j, label=name)
        leg = plt.legend(ncol=6,
                         loc=legend_loc,
                         bbox_to_anchor=legend_bbox,
                         prop={'size':6})
        low = min([min(ratios[x]) for x in non_null])
        high = max([max(ratios[x]) for x in non_null])
        if not isnan(low) and not isnan(high):
            ax.yaxis.set_ticks([low, trunc(((high - low) / 2.0) / 100.) * 100., high])
    else:
        stack = ax.stackplot(timestamps, *r, colors=all_colors)
        ax.set_ylim(0, 100)
        ax.yaxis.set_ticks([0., 50., 100.])
        p = [plt.Rectangle((0, 0), 1, 1, fc=pc.get_facecolor()[0]) for pc in stack]
        leg = plt.legend(p, list(map(suffix, non_null)),
                ncol=6,
                bbox_to_anchor=legend_bbox,
                loc=legend_loc,
                prop={'size':6})
    leg.get_frame().set_alpha(0.5)
    for j in ax.get_xticklabels() + ax.get_yticklabels():
        j.set_fontsize(6)
    if not xaxis:
        xaxis = ax
    #if n >= 2 and not yset and l != -1:
    #    ax.set_ylabel('(% of execution time)')
    #    yset = True
    if n != numplots:
        max_legend = max(len(non_null), max_legend)
    #ax.margins(0, 0)
    n += 1

if len(timestamps) == 1:
    plt.gca().axes.get_xaxis().set_visible(False)

plt.subplots_adjust(hspace=1.5 if max_legend > 6 else 0.9, bottom=0.20,
                    top=0.95)

if args.title:
    #plt.subplot(numplots, 1, 1)
    plt.title(args.title)

if args.output:
    plt.savefig(args.output)
else:
    plt.show()



#!/usr/bin/env python3
# Copyright (c) 2020, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#

# convert toplev output to xlsx files using xlsxwriter
# toplev.py --all --valcsv xv.log --perf-output xp.log -A -a --split-output --per-socket --global --summary \
# --per-core --per-thread -x, -o x.log -I 1000 sleep 10
# tl_xlsx.py --valcsv xv.log --perf xp.log --socket x-socket.log --global x-global.log --core x-core.log --thread x-thread.log x.xlsx
from __future__ import print_function
import sys
import argparse
import re
import collections
import csv
try:
    import xlsxwriter
except ImportError:
    sys.exit("Please install xlswriter with 'pip%s install xlsxwriter'" % sys.version[0])
import gen_level

ap = argparse.ArgumentParser(description="Convert toplev CSV files to xlsx")
ap.add_argument('xlsxfile', help="xlsx output name")
ap.add_argument('--socket', type=argparse.FileType('r'), help="toplev socket csv file", metavar="csvfile")
ap.add_argument('--global', type=argparse.FileType('r'), help="toplev global csv file", dest='global_', metavar="csvfile")
ap.add_argument('--core', type=argparse.FileType('r'), help="toplev core csv file", metavar="csvfile")
ap.add_argument('--program', type=argparse.FileType('r'), help="toplev program csv file", metavar="csvfile")
ap.add_argument('--thread', type=argparse.FileType('r'), help="toplev thread csv file", metavar="csvfile")
ap.add_argument('--add', nargs=2, help="toplev thread generic csv file. Specify csvfile and sheet name as two arguments",
            metavar="name", action="append")
ap.add_argument('--valcsv', type=argparse.FileType('r'), help="toplev valcsv input file", metavar="csvfile")
ap.add_argument('--perf', type=argparse.FileType('r'), help="toplev perf values csv file")
ap.add_argument('--cpuinfo', type=argparse.FileType('r'), help="cpuinfo file")
ap.add_argument('--chart', help="add sheet with plots of normalized sheet. argument is normalied sheet name",
                action="append")
ap.add_argument('--no-summary', help='Do not generate summary charts', action='store_true')
args = ap.parse_args()

workbook = xlsxwriter.Workbook(args.xlsxfile, {'constant_memory': True})
bold = workbook.add_format({'bold': True})
valueformat = workbook.add_format({'num_format': '###,###,###,###,##0.0'})
valueformat_bold = workbook.add_format({'num_format': '###,###,###,###,##0.0',
    'bold': True})
#valueformat.set_num_format(1)

def set_columns(worksheet, c, lengths):
    for col, j in enumerate(c):
        if j == "Value":
            j = " " * 18
        if j == "Description":
            j = "Descr"
        lengths[col] = max(len(j) + 5, lengths[col])
        worksheet.set_column(col, col, lengths[col])

worksheets = {}
rows = {}
headers = {}

def get_worksheet(name):
    if name in worksheets:
        return worksheets[name]
    worksheet = workbook.add_worksheet(name)
    worksheets[name] = worksheet
    rows[name] = 1
    return worksheet

def to_float(n):
    if re.match(r'-?[,.0-9]+$', n):
        n = float(n)
    return n

def create_sheet(name, infh, delimiter=',', version=None):
    lengths = collections.defaultdict(lambda: 0)
    worksheet = get_worksheet(name)
    cf = csv.reader(infh, delimiter=delimiter)
    row = 0
    title = {}
    titlerow = []
    summary = False
    for c in cf:
        if len(c) > 0 and len(c[0]) > 0 and c[0][0] == "#":
            version = c
            continue
        if row == 0:
            title = collections.OrderedDict()
            for i, k in enumerate(c):
                title[k] = i
            titlerow = c
            headers[name] = title
        if row < 10:
            set_columns(worksheet, c, lengths)
        if not summary and len(c) > 0 and c[0] == "SUMMARY":
            if args.no_summary:
                continue
            rows[name] = row
            sname = name + " summary"
            worksheet = get_worksheet(sname)
            set_columns(worksheet, c, lengths)
            worksheet.write_row(0, 0, titlerow)
            headers[sname] = titlerow
            row = rows[sname]
            summary = True
        elif summary and len(c) > 0 and c[0] != "SUMMARY" and c[0][0] != "#":
            worksheet = get_worksheet(name)
            summary = False
            rows[sname] = row
            row = rows[name]
        c = list(map(to_float, c))
        worksheet.write_row(row, 0, c)
        isbn = False
        if "Bottleneck" in title:
            bn = title["Bottleneck"]
            if len(c) > bn and c[bn] == "<==":
                worksheet.write(row, title["Area"], c[title["Area"]], bold)
                worksheet.write(row, title["Value"], c[title["Value"]], bold)
                if "CPUs" in title:
                    worksheet.write(row, title["CPUs"], c[title["CPUs"]], bold)
                isbn = True
                worksheet.write(row, bn, c[bn], bold)
        if "Value" in title and len(c) > title["Value"] and isinstance(c[title["Value"]], float):
            worksheet.write_number(row, title["Value"], c[title["Value"]],
                                   valueformat_bold if isbn else valueformat)
        elif "0" in title:
            num = 0
            while num in title:
                col = title["%d" % num]
                if len(c) > col and re.match(r'[0-9]+', c[col]):
                    worksheet.write_number(row, col, float(c[col]), valueformat)
                num += 1
        row += 1
    return version

ROW_SCALE = 18
MIN_ROWS = 15
GRAPH_ROWS = 15

def gen_chart(source):
    if source not in headers:
        print("source %s for chart not found" % source, file=sys.stderr)
        return

    worksheet = get_worksheet(source + " chart")
    charts = collections.OrderedDict()

    for n, ind in headers[source].items():
        if n == "Timestamp":
            continue
        ns = n.split()
        if len(ns) > 1:
            prefix = ns[0] + " "
            nn = " ".join(ns[1:])
        else:
            prefix = ''
            nn = n

        if gen_level.is_metric(nn):
            chart = workbook.add_chart({'type': 'line'})
            charts[n] = chart
            chart.set_title({'name': n})
        else:
            key = n[:n.rindex('.')] if '.' in n else prefix
            if key not in charts:
                charts[key] = workbook.add_chart(
                        {'type': 'column', 'subtype': 'percent_stacked' })
            chart = charts[key]
            chart.set_title({
                'name': '%s Level %d %s' % (prefix, n.count('.') + 1,
                                            key[key.index(' '):] if ' ' in key else key) })
            chart.set_x_axis({'name': 'Timestamp'})
        chart.add_series({
            'name':        [source, 0, ind],
            'categories':  [source, 1, 0,   rows[source], 0],
            'values':      [source, 1, ind, rows[source], ind]
        })
        chart.set_size({'width': (rows[source] + MIN_ROWS ) * ROW_SCALE})
        chart.set_legend({
            'overlay': True,
            'layout': { 'x': 0.01, 'y': 0.01, 'width': 0.12, 'height': 0.12 },
            'fill': { 'none': True, 'transparency': True }
        })
    row = 1
    for j in charts.keys():
        worksheet.insert_chart('A%d' % row, charts[j])
        row += GRAPH_ROWS

version = None
if args.global_:
    version = create_sheet("global", args.global_, version=version)
if args.socket:
    version = create_sheet("socket", args.socket, version=version)
if args.core:
    version = create_sheet("core", args.core, version=version)
if args.thread:
    version = create_sheet("thread", args.thread, version=version)
if args.program:
    version = create_sheet("prog", args.program, version=version)
if args.add:
    for fn, name in args.add:
        version = create_sheet(name, open(fn), version=version)
if args.chart:
    for cname in args.chart:
        gen_chart(cname)
if args.valcsv:
    create_sheet("event values", args.valcsv)
if args.perf:
    create_sheet("raw perf output", args.perf, delimiter=';')
if args.cpuinfo:
    create_sheet("cpuinfo", args.cpuinfo, delimiter=':')

if version:
    worksheet = workbook.add_worksheet("version")
    worksheet.write_row(0, 0, version)

workbook.close()



# -*- coding: latin-1 -*-
#
# auto generated TopDown/TMA 4 description for Lunarlake Skymont
# Please see http://ark.intel.com for more details on these CPUs.
#
# References:
# http://bit.ly/tma-ispass14
# http://halobates.de/blog/p/262
# https://sites.google.com/site/analysismethods/yasin-pubs
# https://download.01.org/perfmon/
# https://github.com/andikleen/pmu-tools/wiki/toplev-manual
#

# Helpers

print_error = lambda msg: False
version = "4"
base_frequency = -1.0
Memory = 0
Average_Frequency = 0.0
num_cores = 1
num_threads = 1
num_sockets = 1


def handle_error(obj, msg):
    print_error(msg)
    obj.errcount += 1
    obj.val = 0
    obj.thresh = False

def handle_error_metric(obj, msg):
    print_error(msg)
    obj.errcount += 1
    obj.val = 0



# Constants

Pipeline_Width = 8

# Aux. formulas


def CLKS(self, EV, level):
    return EV("CPU_CLK_UNHALTED.CORE", level)

def SLOTS(self, EV, level):
    return Pipeline_Width * CLKS(self, EV, level)

def DurationTimeInSeconds(self, EV, level):
    return EV("interval-ms", 0) / 1000

# Percentage of time that retirement is stalled by the Memory Cluster due to a pipeline stall. See Info.Mem_Exec_Bound
def PCT_Mem_Exec_Bound_Cycles(self, EV, level):
    return 100 * EV("LD_HEAD.ANY_AT_RET", level) / CLKS(self, EV, level)

# Percentage of time that retirement is stalled due to an L1 miss. See Info.Load_Miss_Bound
def PCT_Load_Miss_Bound_Cycles(self, EV, level):
    return 100 * EV("MEM_BOUND_STALLS_LOAD.ALL", level) / CLKS(self, EV, level)

# Percentage of time that retirement is stalled due to a first level data TLB miss
def PCT_DTLB_Miss_Bound_Cycles(self, EV, level):
    return 100 *(EV("LD_HEAD.DTLB_MISS_AT_RET", level) + EV("LD_HEAD.PGWALK_AT_RET", level)) / CLKS(self, EV, level)

# Percentage of time that allocation and retirement is stalled by the Frontend Cluster due to an Ifetch Miss, either Icache or ITLB Miss. See Info.Ifetch_Bound
def PCT_IFetch_Miss_Bound_Cycles(self, EV, level):
    return 100 * EV("MEM_BOUND_STALLS_IFETCH.ALL", level) / CLKS(self, EV, level)

# Instructions Per Cycle
def IPC(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / CLKS(self, EV, level)

# Cycles Per Instruction
def CPI(self, EV, level):
    return CLKS(self, EV, level) / EV("INST_RETIRED.ANY", level)

# Uops Per Instruction
def UPI(self, EV, level):
    return EV("TOPDOWN_RETIRING.ALL", level) / EV("INST_RETIRED.ANY", level)

# Floating Point Operations Per Cycle
def FLOPc(self, EV, level):
    return EV("FP_FLOPS_RETIRED.ALL", level) / CLKS(self, EV, level)

# Instructions per Branch (lower number means higher occurrence rate)
def IpBranch(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_INST_RETIRED.ALL_BRANCHES", level)

# Instruction per (near) call (lower number means higher occurrence rate)
def IpCall(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_INST_RETIRED.NEAR_CALL", level)

# Instructions per Far Branch ( Far Branches apply upon transition from application to operating system, handling interrupts, exceptions) [lower number means higher occurrence rate]
def IpFarBranch(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_INST_RETIRED.FAR_BRANCH:USER", level)

# Instructions per retired Branch Misprediction
def IpMispredict(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_MISP_RETIRED.ALL_BRANCHES", level)

# Instructions per retired conditional Branch Misprediction where the branch was not taken
def IpMisp_Cond_Ntaken(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / (EV("BR_MISP_RETIRED.COND", level) - EV("BR_MISP_RETIRED.COND_TAKEN", level))

# Instructions per retired conditional Branch Misprediction where the branch was taken
def IpMisp_Cond_Taken(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_MISP_RETIRED.COND_TAKEN", level)

# Instructions per retired return Branch Misprediction
def IpMisp_Ret(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_MISP_RETIRED.RETURN", level)

# Instructions per retired indirect call or jump Branch Misprediction
def IpMisp_Indirect(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("BR_MISP_RETIRED.INDIRECT", level)

# Instructions per Floating Point (FP) Operation
def IpFLOP(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("FP_FLOPS_RETIRED.ALL", level)

# Instructions per FP Arithmetic instruction
def IpFPArith(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("FP_INST_RETIRED.ALL", level)

# Instructions per FP Arithmetic Scalar Single-Precision instruction
def IpFPArith_Scalar_SP(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("FP_INST_RETIRED.32B_SP", level)

# Instructions per FP Arithmetic Scalar Double-Precision instruction
def IpFPArith_Scalar_DP(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("FP_INST_RETIRED.64B_DP", level)

# Instructions per FP Arithmetic AVX/SSE 128-bit instruction
def IpFPArith_AVX128(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / (EV("FP_INST_RETIRED.128B_DP", level) + EV("FP_INST_RETIRED.128B_SP", level))

# Instructions per FP Arithmetic AVX 256-bit instruction
def IpFPArith_AVX256(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / (EV("FP_INST_RETIRED.256B_DP", level) + EV("FP_INST_RETIRED.256B_SP", level))

# Percentage of all uops which are microcode ops
def Microcode_Uop_Ratio(self, EV, level):
    return 100 * EV("UOPS_RETIRED.MS", level) / EV("TOPDOWN_RETIRING.ALL", level)

# Percentage of all uops which are FPDiv uops
def FPDiv_Uop_Ratio(self, EV, level):
    return 100 * EV("UOPS_RETIRED.FPDIV", level) / EV("TOPDOWN_RETIRING.ALL", level)

# Percentage of all uops which are IDiv uops
def IDiv_Uop_Ratio(self, EV, level):
    return 100 * EV("UOPS_RETIRED.IDIV", level) / EV("TOPDOWN_RETIRING.ALL", level)

# Percentage of all uops which are x87 uops
def X87_Uop_Ratio(self, EV, level):
    return 100 * EV("UOPS_RETIRED.X87", level) / EV("TOPDOWN_RETIRING.ALL", level)

# Instructions per Load
def IpLoad(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("MEM_UOPS_RETIRED.ALL_LOADS", level)

# Instructions per Store
def IpStore(self, EV, level):
    return EV("INST_RETIRED.ANY", level) / EV("MEM_UOPS_RETIRED.ALL_STORES", level)

# Ratio of mem load uops to all uops
def MemLoad_Ratio(self, EV, level):
    return 1000 * EV("MEM_UOPS_RETIRED.ALL_LOADS", level) / EV("TOPDOWN_RETIRING.ALL", level)

# Percentage of total non-speculative loads that are splits
def Load_Splits_Ratio(self, EV, level):
    return 100 * EV("MEM_UOPS_RETIRED.SPLIT_LOADS", level) / EV("MEM_UOPS_RETIRED.ALL_LOADS", level)

# Percentage of total non-speculative loads that perform one or more locks
def Load_Locks_Ratio(self, EV, level):
    return 100 * EV("MEM_UOPS_RETIRED.LOCK_LOADS", level) / EV("MEM_UOPS_RETIRED.ALL_LOADS", level)

# Percentage of total non-speculative loads with a store forward or unknown store address block
def PCT_Loads_with_StoreFwdBlk(self, EV, level):
    return 100 * EV("LD_BLOCKS.DATA_UNKNOWN", level) / EV("MEM_UOPS_RETIRED.ALL_LOADS", level)

# Percentage of total non-speculative loads with an address aliasing block
def PCT_Loads_with_AdressAliasing(self, EV, level):
    return 100 * EV("LD_BLOCKS.ADDRESS_ALIAS", level) / EV("MEM_UOPS_RETIRED.ALL_LOADS", level)

# Percentage of Memory Execution Bound due to a second level TLB miss
def PCT_LoadHead_with_STLBHit(self, EV, level):
    return 100 * EV("LD_HEAD.DTLB_MISS_AT_RET", level) / EV("LD_HEAD.ANY_AT_RET", level)

# Percentage of Memory Execution Bound due to a pagewalk
def PCT_LoadHead_with_Pagewalk(self, EV, level):
    return 100 * EV("LD_HEAD.PGWALK_AT_RET", level) / EV("LD_HEAD.ANY_AT_RET", level)

# Percentage of Memory Execution Bound due to a store forward address match
def PCT_LoadHead_with_StoreFwding(self, EV, level):
    return 100 * EV("LD_HEAD.ST_ADDR_AT_RET", level) / EV("LD_HEAD.ANY_AT_RET", level)

# Percentage of Memory Execution Bound due to other block cases, such as pipeline conflicts, fences, etc
def PCT_LoadHead_with_OtherPipelineBlks(self, EV, level):
    return 100 * EV("LD_HEAD.OTHER_AT_RET", level) / EV("LD_HEAD.ANY_AT_RET", level)

# Percentage of Memory Execution Bound due to a first level data cache miss
def PCT_LoadHead_with_L1miss(self, EV, level):
    return 100 * EV("LD_HEAD.L1_MISS_AT_RET", level) / EV("LD_HEAD.ANY_AT_RET", level)

# Counts the number of cycles the core is stalled due to store buffer full
def Store_Bound(self, EV, level):
    return 100 *(EV("MEM_SCHEDULER_BLOCK.ST_BUF", level) / EV("MEM_SCHEDULER_BLOCK.ALL", level)) * (EV("TOPDOWN_BE_BOUND.MEM_SCHEDULER", level) / SLOTS(self, EV, level))

# Counts the number of cycles that the oldest load of the load buffer is stalled at retirement
def Load_Bound(self, EV, level):
    return 100 *(EV("LD_HEAD.L1_BOUND_AT_RET", level) + EV("MEM_BOUND_STALLS_LOAD.ALL", level)) / CLKS(self, EV, level)

# Counts the number of cycles that the oldest load of the load buffer is stalled at retirement due to a pipeline block
def L1_Bound(self, EV, level):
    return 100 * EV("LD_HEAD.L1_BOUND_AT_RET", level) / CLKS(self, EV, level)

# Percentage of memory bound stalls where retirement is stalled due to an L1 miss that hit the L2
def PCT_LoadMissBound_with_L2Hit(self, EV, level):
    return 100 * EV("MEM_BOUND_STALLS_LOAD.L2_HIT", level) / EV("MEM_BOUND_STALLS_LOAD.ALL", level)

# Percentage of memory bound stalls where retirement is stalled due to an L1 miss that subsequently misses in the L2
def PCT_LoadMissBound_with_L2Miss(self, EV, level):
    return 100 * EV("MEM_BOUND_STALLS_LOAD.L2_MISS", level) / EV("MEM_BOUND_STALLS_LOAD.ALL", level)

# Percentage of ifetch miss bound stalls, where the ifetch miss hits in the L2
def PCT_IfetchMissBound_with_L2Hit(self, EV, level):
    return 100 * EV("MEM_BOUND_STALLS_IFETCH.L2_HIT", level) / EV("MEM_BOUND_STALLS_IFETCH.ALL", level)

# Percentage of ifetch miss bound stalls, where the ifetch miss doesn't hit in the L2
def PCT_IfetchMissBound_with_L2Miss(self, EV, level):
    return 100 * EV("MEM_BOUND_STALLS_IFETCH.L2_MISS", level) / EV("MEM_BOUND_STALLS_IFETCH.ALL", level)

# Ratio of all branches which mispredict
def Branch_Mispredict_Ratio(self, EV, level):
    return EV("BR_MISP_RETIRED.ALL_BRANCHES", level) / EV("BR_INST_RETIRED.ALL_BRANCHES", level)

# Ratio between Mispredicted branches and unknown branches
def Branch_Mispredict_to_Unknown_Branch_Ratio(self, EV, level):
    return EV("BR_MISP_RETIRED.ALL_BRANCHES", level) / EV("BACLEARS.ANY", level)

# Counts the number of machine clears relative to thousands of instructions retired, due to memory disambiguation
def Machine_Clears_Disamb_PKI(self, EV, level):
    return 1000 * EV("MACHINE_CLEARS.DISAMBIGUATION", level) / EV("INST_RETIRED.ANY", level)

# Counts the number of machine clears relative to thousands of instructions retired, due to floating point assists
def Machine_Clears_FP_Assist_PKI(self, EV, level):
    return 1000 * EV("MACHINE_CLEARS.FP_ASSIST", level) / EV("INST_RETIRED.ANY", level)

# Counts the number of machine clears relative to thousands of instructions retired, due to memory ordering
def Machine_Clears_Monuke_PKI(self, EV, level):
    return 1000 * EV("MACHINE_CLEARS.MEMORY_ORDERING", level) / EV("INST_RETIRED.ANY", level)

# Counts the number of machine clears relative to thousands of instructions retired, due to page faults
def Machine_Clears_Page_Fault_PKI(self, EV, level):
    return 1000 * EV("MACHINE_CLEARS.PAGE_FAULT", level) / EV("INST_RETIRED.ANY", level)

# Counts the number of machine clears relative to thousands of instructions retired, due to memory renaming
def Machine_Clears_MRN_PKI(self, EV, level):
    return 1000 * EV("MACHINE_CLEARS.MRN_NUKE", level) / EV("INST_RETIRED.ANY", level)

# Percentage of time that allocation is stalled due to store buffer full
def PCT_Store_Buffer_Stall_Cycles(self, EV, level):
    return 100 * EV("MEM_SCHEDULER_BLOCK.ST_BUF", level) / CLKS(self, EV, level)

# Percentage of time that allocation is stalled due to load buffer full
def PCT_Load_Buffer_Stall_Cycles(self, EV, level):
    return 100 * EV("MEM_SCHEDULER_BLOCK.LD_BUF", level) / CLKS(self, EV, level)

# Percentage of time that allocation is stalled due to memory reservation stations full
def PCT_Mem_RSV_Stall_Cycles(self, EV, level):
    return 100 * EV("MEM_SCHEDULER_BLOCK.RSV", level) / CLKS(self, EV, level)

# Percentage of time that the core is stalled due to a TPAUSE or UMWAIT instruction 
def PCT_Tpause_Cycles(self, EV, level):
    return 100 * EV("SERIALIZATION.C01_MS_SCB", level) / SLOTS(self, EV, level)

# Average Frequency Utilization relative nominal frequency
def Turbo_Utilization(self, EV, level):
    return CLKS(self, EV, level) / EV("CPU_CLK_UNHALTED.REF_TSC", level)

# Fraction of cycles spent in Kernel mode
def Kernel_Utilization(self, EV, level):
    return EV("CPU_CLK_UNHALTED.CORE_P:sup", level) / CLKS(self, EV, level)

# Average CPU Utilization
def CPU_Utilization(self, EV, level):
    return EV("CPU_CLK_UNHALTED.REF_TSC", level) / EV("msr/tsc/", 0)

# Giga Floating Point Operations Per Second. Aggregate across all supported options of: FP precisions, scalar and vector instructions, vector-width
def GFLOPs(self, EV, level):
    return EV("FP_FLOPS_RETIRED.ALL", level) / (EV("interval-s", 0) * 1000000000 )

# Event groups


class Frontend_Bound:
    name = "Frontend_Bound"
    domain = "Slots"
    area = "FE"
    level = 1
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.ALL", 1) / SLOTS(self, EV, 1)
            self.thresh = (self.val > 0.20)
        except ZeroDivisionError:
            handle_error(self, "Frontend_Bound zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not consumed by
the backend due to frontend stalls."""


class IFetch_Latency:
    name = "IFetch_Latency"
    domain = "Slots"
    area = "FE"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.FRONTEND_LATENCY", 2) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.15) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "IFetch_Latency zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not delivered by
the frontend due to frontend latency restrictions due to
icache misses, itlb misses, branch detection, and resteer
limitations."""


class ICache_Misses:
    name = "ICache_Misses"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.ICACHE", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "ICache_Misses zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not delivered by
the frontend due to instruction cache misses."""


class ITLB_Misses:
    name = "ITLB_Misses"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.ITLB_MISS", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "ITLB_Misses zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not delivered by
the frontend due to Instruction Table Lookaside Buffer
(ITLB) misses."""


class Branch_Detect:
    name = "Branch_Detect"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.BRANCH_DETECT", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Branch_Detect zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not delivered by
the frontend due to BACLEARS, which occurs when the Branch
Target Buffer (BTB) prediction or lack thereof, was
corrected by a later branch predictor in the frontend.
Includes BACLEARS due to all branch types including
conditional and unconditional jumps, returns, and indirect
branches."""


class Branch_Resteer:
    name = "Branch_Resteer"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.BRANCH_RESTEER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Branch_Resteer zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not delivered by
the frontend due to BTCLEARS, which occurs when the Branch
Target Buffer (BTB) predicts a taken branch."""


class IFetch_Bandwidth:
    name = "IFetch_Bandwidth"
    domain = "Slots"
    area = "FE"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.FRONTEND_BANDWIDTH", 2) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.10) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "IFetch_Bandwidth zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not delivered by
the frontend due to frontend bandwidth restrictions due to
decode, predecode, cisc, and other limitations."""


class Cisc:
    name = "Cisc"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.CISC", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Cisc zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not delivered by
the frontend due to the microcode sequencer (MS)."""


class Decode:
    name = "Decode"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.DECODE", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Decode zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not delivered by
the frontend due to decode stalls."""


class Predecode:
    name = "Predecode"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.PREDECODE", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Predecode zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not delivered by
the frontend due to wrong predecodes."""


class Other_FB:
    name = "Other_FB"
    domain = "Slots"
    area = "FE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_FE_BOUND.OTHER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Other_FB zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not delivered by
the frontend due to other common frontend stalls not
categorized."""


class Bad_Speculation:
    name = "Bad_Speculation"
    domain = "Slots"
    area = "BAD"
    level = 1
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BAD_SPECULATION.ALL_P", 1) / SLOTS(self, EV, 1)
            self.thresh = (self.val > 0.15)
        except ZeroDivisionError:
            handle_error(self, "Bad_Speculation zero division")
        return self.val
    desc = """
Counts the total number of issue slots that were not
consumed by the backend because allocation is stalled due to
a mispredicted jump or a machine clear. Only issue slots
wasted due to fast nukes such as memory ordering nukes are
counted. Other nukes are not accounted for. Counts all issue
slots blocked during this recovery window including relevant
microcode flows and while uops are not yet available in the
instruction queue (IQ). Also includes the issue slots that
were consumed by the backend but were thrown away because
they were younger than the mispredict or machine clear."""


class Branch_Mispredicts:
    name = "Branch_Mispredicts"
    domain = "Slots"
    area = "BAD"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BAD_SPECULATION.MISPREDICT", 2) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Branch_Mispredicts zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not consumed by
the backend due to branch mispredicts"""


class Machine_Clears:
    name = "Machine_Clears"
    domain = "Slots"
    area = "BAD"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BAD_SPECULATION.MACHINE_CLEARS", 2) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Machine_Clears zero division")
        return self.val
    desc = """
Counts the total number of issue slots that were not
consumed by the backend because allocation is stalled due to
a machine clear (nuke) of any kind including memory ordering
and memory disambiguation"""


class Nuke:
    name = "Nuke"
    domain = "Slots"
    area = "BAD"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BAD_SPECULATION.NUKE", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Nuke zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not consumed by
the backend due to a machine clear that requires the use of
microcode (slow nuke)"""


class Fast_Nuke:
    name = "Fast_Nuke"
    domain = "Slots"
    area = "BAD"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BAD_SPECULATION.FASTNUKE", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.05) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Fast_Nuke zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not consumed by
the backend due to a machine clear that does not require the
use of microcode, classified as a fast nuke, due to memory
ordering, memory disambiguation and memory renaming"""


class Backend_Bound:
    name = "Backend_Bound"
    domain = "Slots"
    area = "BE"
    level = 1
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.ALL_P", 1) / SLOTS(self, EV, 1)
            self.thresh = (self.val > 0.10)
        except ZeroDivisionError:
            handle_error(self, "Backend_Bound zero division")
        return self.val
    desc = """
Counts the total number of issue slots that were not
consumed by the backend due to backend stalls. Note that
uops must be available for consumption in order for this
event to count. If a uop is not available (IQ is empty),
this event will not count"""


class Core_Bound:
    name = "Core_Bound"
    domain = "Slots"
    area = "BE"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.ALLOC_RESTRICTIONS", 2) / SLOTS(self, EV, 2)
            self.thresh = (self.val > 0.10) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Core_Bound zero division")
        return self.val
    desc = """
Counts the number of cycles due to backend bound stalls that
are bounded by core restrictions and not attributed to an
outstanding load or stores, or resource limitation"""


class Allocation_Restriction:
    name = "Allocation_Restriction"
    domain = "Slots"
    area = "BE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.ALLOC_RESTRICTIONS", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Allocation_Restriction zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not consumed by
the backend due to certain allocation restrictions"""


class Resource_Bound:
    name = "Resource_Bound"
    domain = "Slots"
    area = "BE"
    level = 2
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = ((EV("TOPDOWN_BE_BOUND.ALL_P", 2) / SLOTS(self, EV, 2)) - self.Core_Bound.compute(EV))
            self.thresh = (self.val > 0.20) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Resource_Bound zero division")
        return self.val
    desc = """
Counts the number of cycles the core is stalled due to a
resource limitation"""


class Mem_Scheduler:
    name = "Mem_Scheduler"
    domain = "Slots"
    area = "BE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.MEM_SCHEDULER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Mem_Scheduler zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not consumed by
the backend due to memory reservation stalls in which a
scheduler is not able to accept uops"""


class Non_Mem_Scheduler:
    name = "Non_Mem_Scheduler"
    domain = "Slots"
    area = "BE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.NON_MEM_SCHEDULER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Non_Mem_Scheduler zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not consumed by
the backend due to IEC or FPC RAT stalls, which can be due
to FIQ or IEC reservation stalls in which the integer,
floating point or SIMD scheduler is not able to accept uops"""


class Register:
    name = "Register"
    domain = "Slots"
    area = "BE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.REGISTER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Register zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not consumed by
the backend due to the physical register file unable to
accept an entry (marble stalls)"""


class Reorder_Buffer:
    name = "Reorder_Buffer"
    domain = "Slots"
    area = "BE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.REORDER_BUFFER", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Reorder_Buffer zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not consumed by
the backend due to the reorder buffer being full (ROB
stalls)"""


class Serialization:
    name = "Serialization"
    domain = "Slots"
    area = "BE"
    level = 3
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_BE_BOUND.SERIALIZATION", 3) / SLOTS(self, EV, 3)
            self.thresh = (self.val > 0.10) and self.parent.thresh
        except ZeroDivisionError:
            handle_error(self, "Serialization zero division")
        return self.val
    desc = """
Counts the number of issue slots that were not consumed by
the backend due to scoreboards from the instruction queue
(IQ), jump execution unit (JEU), or microcode sequencer (MS)"""


class Retiring:
    name = "Retiring"
    domain = "Slots"
    area = "RET"
    level = 1
    htoff = False
    sample = []
    errcount = 0
    sibling = None
    metricgroup = frozenset([])
    maxval = None
    def compute(self, EV):
        try:
            self.val = EV("TOPDOWN_RETIRING.ALL", 1) / SLOTS(self, EV, 1)
            self.thresh = (self.val > 0.75)
        except ZeroDivisionError:
            handle_error(self, "Retiring zero division")
        return self.val
    desc = """
Counts the number of issue slots that result in retirement
slots"""


class Metric_PCT_Mem_Exec_Bound_Cycles:
    name = "PCT_Mem_Exec_Bound_Cycles"
    domain = "Cycles"
    maxval = 0
    errcount = 0
    area = "Info.Bottleneck"
    metricgroup = frozenset(['Mem_Exec'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_Mem_Exec_Bound_Cycles(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_Mem_Exec_Bound_Cycles zero division")
    desc = """
Percentage of time that retirement is stalled by the Memory
Cluster due to a pipeline stall. See Info.Mem_Exec_Bound"""


class Metric_PCT_Load_Miss_Bound_Cycles:
    name = "PCT_Load_Miss_Bound_Cycles"
    domain = "Cycles"
    maxval = 0
    errcount = 0
    area = "Info.Bottleneck"
    metricgroup = frozenset(['Load_Store_Miss'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_Load_Miss_Bound_Cycles(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_Load_Miss_Bound_Cycles zero division")
    desc = """
Percentage of time that retirement is stalled due to an L1
miss. See Info.Load_Miss_Bound"""


class Metric_PCT_DTLB_Miss_Bound_Cycles:
    name = "PCT_DTLB_Miss_Bound_Cycles"
    domain = "Cycles"
    maxval = 0
    errcount = 0
    area = "Info.Bottleneck"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_DTLB_Miss_Bound_Cycles(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_DTLB_Miss_Bound_Cycles zero division")
    desc = """
Percentage of time that retirement is stalled due to a first
level data TLB miss"""


class Metric_PCT_IFetch_Miss_Bound_Cycles:
    name = "PCT_IFetch_Miss_Bound_Cycles"
    domain = "Cycles"
    maxval = 0
    errcount = 0
    area = "Info.Bottleneck"
    metricgroup = frozenset(['Ifetch'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_IFetch_Miss_Bound_Cycles(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_IFetch_Miss_Bound_Cycles zero division")
    desc = """
Percentage of time that allocation and retirement is stalled
by the Frontend Cluster due to an Ifetch Miss, either Icache
or ITLB Miss. See Info.Ifetch_Bound"""


class Metric_IPC:
    name = "IPC"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Core"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IPC(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IPC zero division")
    desc = """
Instructions Per Cycle"""


class Metric_CPI:
    name = "CPI"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Core"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = CPI(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "CPI zero division")
    desc = """
Cycles Per Instruction"""


class Metric_UPI:
    name = "UPI"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Core"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = UPI(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "UPI zero division")
    desc = """
Uops Per Instruction"""


class Metric_FLOPc:
    name = "FLOPc"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Core"
    metricgroup = frozenset(['Flops'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = FLOPc(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "FLOPc zero division")
    desc = """
Floating Point Operations Per Cycle"""


class Metric_IpBranch:
    name = "IpBranch"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Br_Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpBranch(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpBranch zero division")
    desc = """
Instructions per Branch (lower number means higher
occurrence rate)"""


class Metric_IpCall:
    name = "IpCall"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Br_Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpCall(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpCall zero division")
    desc = """
Instruction per (near) call (lower number means higher
occurrence rate)"""


class Metric_IpFarBranch:
    name = "IpFarBranch"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Br_Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpFarBranch(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpFarBranch zero division")
    desc = """
Instructions per Far Branch ( Far Branches apply upon
transition from application to operating system, handling
interrupts, exceptions) [lower number means higher
occurrence rate]"""


class Metric_IpMispredict:
    name = "IpMispredict"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Br_Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpMispredict(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpMispredict zero division")
    desc = """
Instructions per retired Branch Misprediction"""


class Metric_IpMisp_Cond_Ntaken:
    name = "IpMisp_Cond_Ntaken"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Br_Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpMisp_Cond_Ntaken(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpMisp_Cond_Ntaken zero division")
    desc = """
Instructions per retired conditional Branch Misprediction
where the branch was not taken"""


class Metric_IpMisp_Cond_Taken:
    name = "IpMisp_Cond_Taken"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Br_Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpMisp_Cond_Taken(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpMisp_Cond_Taken zero division")
    desc = """
Instructions per retired conditional Branch Misprediction
where the branch was taken"""


class Metric_IpMisp_Ret:
    name = "IpMisp_Ret"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Br_Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpMisp_Ret(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpMisp_Ret zero division")
    desc = """
Instructions per retired return Branch Misprediction"""


class Metric_IpMisp_Indirect:
    name = "IpMisp_Indirect"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Br_Inst_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpMisp_Indirect(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpMisp_Indirect zero division")
    desc = """
Instructions per retired indirect call or jump Branch
Misprediction"""


class Metric_IpFLOP:
    name = "IpFLOP"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Arith_Inst_Mix"
    metricgroup = frozenset(['Flops'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpFLOP(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpFLOP zero division")
    desc = """
Instructions per Floating Point (FP) Operation"""


class Metric_IpFPArith:
    name = "IpFPArith"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Arith_Inst_Mix"
    metricgroup = frozenset(['Flops'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpFPArith(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpFPArith zero division")
    desc = """
Instructions per FP Arithmetic instruction"""


class Metric_IpFPArith_Scalar_SP:
    name = "IpFPArith_Scalar_SP"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Arith_Inst_Mix"
    metricgroup = frozenset(['Flops'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpFPArith_Scalar_SP(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpFPArith_Scalar_SP zero division")
    desc = """
Instructions per FP Arithmetic Scalar Single-Precision
instruction"""


class Metric_IpFPArith_Scalar_DP:
    name = "IpFPArith_Scalar_DP"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Arith_Inst_Mix"
    metricgroup = frozenset(['Flops'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpFPArith_Scalar_DP(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpFPArith_Scalar_DP zero division")
    desc = """
Instructions per FP Arithmetic Scalar Double-Precision
instruction"""


class Metric_IpFPArith_AVX128:
    name = "IpFPArith_AVX128"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Arith_Inst_Mix"
    metricgroup = frozenset(['Flops'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpFPArith_AVX128(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpFPArith_AVX128 zero division")
    desc = """
Instructions per FP Arithmetic AVX/SSE 128-bit instruction"""


class Metric_IpFPArith_AVX256:
    name = "IpFPArith_AVX256"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Arith_Inst_Mix"
    metricgroup = frozenset(['Flops'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpFPArith_AVX256(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpFPArith_AVX256 zero division")
    desc = """
Instructions per FP Arithmetic AVX 256-bit instruction"""


class Metric_Microcode_Uop_Ratio:
    name = "Microcode_Uop_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Uop_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Microcode_Uop_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Microcode_Uop_Ratio zero division")
    desc = """
Percentage of all uops which are microcode ops"""


class Metric_FPDiv_Uop_Ratio:
    name = "FPDiv_Uop_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Uop_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = FPDiv_Uop_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "FPDiv_Uop_Ratio zero division")
    desc = """
Percentage of all uops which are FPDiv uops"""


class Metric_IDiv_Uop_Ratio:
    name = "IDiv_Uop_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Uop_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IDiv_Uop_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IDiv_Uop_Ratio zero division")
    desc = """
Percentage of all uops which are IDiv uops"""


class Metric_X87_Uop_Ratio:
    name = "X87_Uop_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Uop_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = X87_Uop_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "X87_Uop_Ratio zero division")
    desc = """
Percentage of all uops which are x87 uops"""


class Metric_IpLoad:
    name = "IpLoad"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpLoad(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpLoad zero division")
    desc = """
Instructions per Load"""


class Metric_IpStore:
    name = "IpStore"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = IpStore(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "IpStore zero division")
    desc = """
Instructions per Store"""


class Metric_MemLoad_Ratio:
    name = "MemLoad_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = MemLoad_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "MemLoad_Ratio zero division")
    desc = """
Ratio of mem load uops to all uops"""


class Metric_Load_Splits_Ratio:
    name = "Load_Splits_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Load_Splits_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Load_Splits_Ratio zero division")
    desc = """
Percentage of total non-speculative loads that are splits"""


class Metric_Load_Locks_Ratio:
    name = "Load_Locks_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Mix"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Load_Locks_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Load_Locks_Ratio zero division")
    desc = """
Percentage of total non-speculative loads that perform one
or more locks"""


class Metric_PCT_Loads_with_StoreFwdBlk:
    name = "PCT_Loads_with_StoreFwdBlk"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Exec_Blocks"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_Loads_with_StoreFwdBlk(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_Loads_with_StoreFwdBlk zero division")
    desc = """
Percentage of total non-speculative loads with a store
forward or unknown store address block"""


class Metric_PCT_Loads_with_AdressAliasing:
    name = "PCT_Loads_with_AdressAliasing"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Exec_Blocks"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_Loads_with_AdressAliasing(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_Loads_with_AdressAliasing zero division")
    desc = """
Percentage of total non-speculative loads with an address
aliasing block"""


class Metric_PCT_LoadHead_with_STLBHit:
    name = "PCT_LoadHead_with_STLBHit"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Exec_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_LoadHead_with_STLBHit(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_LoadHead_with_STLBHit zero division")
    desc = """
Percentage of Memory Execution Bound due to a second level
TLB miss"""


class Metric_PCT_LoadHead_with_Pagewalk:
    name = "PCT_LoadHead_with_Pagewalk"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Exec_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_LoadHead_with_Pagewalk(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_LoadHead_with_Pagewalk zero division")
    desc = """
Percentage of Memory Execution Bound due to a pagewalk"""


class Metric_PCT_LoadHead_with_StoreFwding:
    name = "PCT_LoadHead_with_StoreFwding"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Exec_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_LoadHead_with_StoreFwding(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_LoadHead_with_StoreFwding zero division")
    desc = """
Percentage of Memory Execution Bound due to a store forward
address match"""


class Metric_PCT_LoadHead_with_OtherPipelineBlks:
    name = "PCT_LoadHead_with_OtherPipelineBlks"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Exec_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_LoadHead_with_OtherPipelineBlks(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_LoadHead_with_OtherPipelineBlks zero division")
    desc = """
Percentage of Memory Execution Bound due to other block
cases, such as pipeline conflicts, fences, etc"""


class Metric_PCT_LoadHead_with_L1miss:
    name = "PCT_LoadHead_with_L1miss"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Mem_Exec_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_LoadHead_with_L1miss(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_LoadHead_with_L1miss zero division")
    desc = """
Percentage of Memory Execution Bound due to a first level
data cache miss"""


class Metric_Store_Bound:
    name = "Store_Bound"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Load_Store_Bound"
    metricgroup = frozenset(['load_store_bound'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Store_Bound(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Store_Bound zero division")
    desc = """
Counts the number of cycles the core is stalled due to store
buffer full"""


class Metric_Load_Bound:
    name = "Load_Bound"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Load_Store_Bound"
    metricgroup = frozenset(['load_store_bound'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Load_Bound(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Load_Bound zero division")
    desc = """
Counts the number of cycles that the oldest load of the load
buffer is stalled at retirement"""


class Metric_L1_Bound:
    name = "L1_Bound"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Load_Store_Bound"
    metricgroup = frozenset(['load_store_bound'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = L1_Bound(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "L1_Bound zero division")
    desc = """
Counts the number of cycles that the oldest load of the load
buffer is stalled at retirement due to a pipeline block"""


class Metric_PCT_LoadMissBound_with_L2Hit:
    name = "PCT_LoadMissBound_with_L2Hit"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Load_Miss_Bound"
    metricgroup = frozenset(['load_store_bound'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_LoadMissBound_with_L2Hit(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_LoadMissBound_with_L2Hit zero division")
    desc = """
Percentage of memory bound stalls where retirement is
stalled due to an L1 miss that hit the L2"""


class Metric_PCT_LoadMissBound_with_L2Miss:
    name = "PCT_LoadMissBound_with_L2Miss"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Load_Miss_Bound"
    metricgroup = frozenset(['load_store_bound'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_LoadMissBound_with_L2Miss(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_LoadMissBound_with_L2Miss zero division")
    desc = """
Percentage of memory bound stalls where retirement is
stalled due to an L1 miss that subsequently misses in the L2"""


class Metric_PCT_IfetchMissBound_with_L2Hit:
    name = "PCT_IfetchMissBound_with_L2Hit"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Ifetch_Miss_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_IfetchMissBound_with_L2Hit(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_IfetchMissBound_with_L2Hit zero division")
    desc = """
Percentage of ifetch miss bound stalls, where the ifetch
miss hits in the L2"""


class Metric_PCT_IfetchMissBound_with_L2Miss:
    name = "PCT_IfetchMissBound_with_L2Miss"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Ifetch_Miss_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_IfetchMissBound_with_L2Miss(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_IfetchMissBound_with_L2Miss zero division")
    desc = """
Percentage of ifetch miss bound stalls, where the ifetch
miss doesn't hit in the L2"""


class Metric_Branch_Mispredict_Ratio:
    name = "Branch_Mispredict_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Br_Mispredict_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Branch_Mispredict_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Branch_Mispredict_Ratio zero division")
    desc = """
Ratio of all branches which mispredict"""


class Metric_Branch_Mispredict_to_Unknown_Branch_Ratio:
    name = "Branch_Mispredict_to_Unknown_Branch_Ratio"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Br_Mispredict_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Branch_Mispredict_to_Unknown_Branch_Ratio(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Branch_Mispredict_to_Unknown_Branch_Ratio zero division")
    desc = """
Ratio between Mispredicted branches and unknown branches"""


class Metric_Machine_Clears_Disamb_PKI:
    name = "Machine_Clears_Disamb_PKI"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Machine_Clear_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Machine_Clears_Disamb_PKI(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Machine_Clears_Disamb_PKI zero division")
    desc = """
Counts the number of machine clears relative to thousands of
instructions retired, due to memory disambiguation"""


class Metric_Machine_Clears_FP_Assist_PKI:
    name = "Machine_Clears_FP_Assist_PKI"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Machine_Clear_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Machine_Clears_FP_Assist_PKI(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Machine_Clears_FP_Assist_PKI zero division")
    desc = """
Counts the number of machine clears relative to thousands of
instructions retired, due to floating point assists"""


class Metric_Machine_Clears_Monuke_PKI:
    name = "Machine_Clears_Monuke_PKI"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Machine_Clear_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Machine_Clears_Monuke_PKI(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Machine_Clears_Monuke_PKI zero division")
    desc = """
Counts the number of machine clears relative to thousands of
instructions retired, due to memory ordering"""


class Metric_Machine_Clears_Page_Fault_PKI:
    name = "Machine_Clears_Page_Fault_PKI"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Machine_Clear_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Machine_Clears_Page_Fault_PKI(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Machine_Clears_Page_Fault_PKI zero division")
    desc = """
Counts the number of machine clears relative to thousands of
instructions retired, due to page faults"""


class Metric_Machine_Clears_MRN_PKI:
    name = "Machine_Clears_MRN_PKI"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Machine_Clear_Bound"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Machine_Clears_MRN_PKI(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Machine_Clears_MRN_PKI zero division")
    desc = """
Counts the number of machine clears relative to thousands of
instructions retired, due to memory renaming"""


class Metric_PCT_Store_Buffer_Stall_Cycles:
    name = "PCT_Store_Buffer_Stall_Cycles"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Buffer_Stalls"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_Store_Buffer_Stall_Cycles(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_Store_Buffer_Stall_Cycles zero division")
    desc = """
Percentage of time that allocation is stalled due to store
buffer full"""


class Metric_PCT_Load_Buffer_Stall_Cycles:
    name = "PCT_Load_Buffer_Stall_Cycles"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Buffer_Stalls"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_Load_Buffer_Stall_Cycles(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_Load_Buffer_Stall_Cycles zero division")
    desc = """
Percentage of time that allocation is stalled due to load
buffer full"""


class Metric_PCT_Mem_RSV_Stall_Cycles:
    name = "PCT_Mem_RSV_Stall_Cycles"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Buffer_Stalls"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_Mem_RSV_Stall_Cycles(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_Mem_RSV_Stall_Cycles zero division")
    desc = """
Percentage of time that allocation is stalled due to memory
reservation stations full"""


class Metric_PCT_Tpause_Cycles:
    name = "PCT_Tpause_Cycles"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.Serialization "
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = PCT_Tpause_Cycles(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "PCT_Tpause_Cycles zero division")
    desc = """
Percentage of time that the core is stalled due to a TPAUSE
or UMWAIT instruction"""


class Metric_Turbo_Utilization:
    name = "Turbo_Utilization"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.System"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Turbo_Utilization(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Turbo_Utilization zero division")
    desc = """
Average Frequency Utilization relative nominal frequency"""


class Metric_Kernel_Utilization:
    name = "Kernel_Utilization"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.System"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = Kernel_Utilization(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "Kernel_Utilization zero division")
    desc = """
Fraction of cycles spent in Kernel mode"""


class Metric_CPU_Utilization:
    name = "CPU_Utilization"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.System"
    metricgroup = frozenset([])
    sibling = None

    def compute(self, EV):
        try:
            self.val = CPU_Utilization(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "CPU_Utilization zero division")
    desc = """
Average CPU Utilization"""


class Metric_GFLOPs:
    name = "GFLOPs"
    domain = ""
    maxval = 0
    errcount = 0
    area = "Info.System"
    metricgroup = frozenset(['Flops'])
    sibling = None

    def compute(self, EV):
        try:
            self.val = GFLOPs(self, EV, 0)
            self.thresh = True
        except ZeroDivisionError:
            handle_error_metric(self, "GFLOPs zero division")
    desc = """
Giga Floating Point Operations Per Second. Aggregate across
all supported options of: FP precisions, scalar and vector
instructions, vector-width"""


# Schedule



class Setup:
    def __init__(self, r):
        o = dict()
        n = Frontend_Bound() ; r.run(n) ; o["Frontend_Bound"] = n
        n = IFetch_Latency() ; r.run(n) ; o["IFetch_Latency"] = n
        n = ICache_Misses() ; r.run(n) ; o["ICache_Misses"] = n
        n = ITLB_Misses() ; r.run(n) ; o["ITLB_Misses"] = n
        n = Branch_Detect() ; r.run(n) ; o["Branch_Detect"] = n
        n = Branch_Resteer() ; r.run(n) ; o["Branch_Resteer"] = n
        n = IFetch_Bandwidth() ; r.run(n) ; o["IFetch_Bandwidth"] = n
        n = Cisc() ; r.run(n) ; o["Cisc"] = n
        n = Decode() ; r.run(n) ; o["Decode"] = n
        n = Predecode() ; r.run(n) ; o["Predecode"] = n
        n = Other_FB() ; r.run(n) ; o["Other_FB"] = n
        n = Bad_Speculation() ; r.run(n) ; o["Bad_Speculation"] = n
        n = Branch_Mispredicts() ; r.run(n) ; o["Branch_Mispredicts"] = n
        n = Machine_Clears() ; r.run(n) ; o["Machine_Clears"] = n
        n = Nuke() ; r.run(n) ; o["Nuke"] = n
        n = Fast_Nuke() ; r.run(n) ; o["Fast_Nuke"] = n
        n = Backend_Bound() ; r.run(n) ; o["Backend_Bound"] = n
        n = Core_Bound() ; r.run(n) ; o["Core_Bound"] = n
        n = Allocation_Restriction() ; r.run(n) ; o["Allocation_Restriction"] = n
        n = Resource_Bound() ; r.run(n) ; o["Resource_Bound"] = n
        n = Mem_Scheduler() ; r.run(n) ; o["Mem_Scheduler"] = n
        n = Non_Mem_Scheduler() ; r.run(n) ; o["Non_Mem_Scheduler"] = n
        n = Register() ; r.run(n) ; o["Register"] = n
        n = Reorder_Buffer() ; r.run(n) ; o["Reorder_Buffer"] = n
        n = Serialization() ; r.run(n) ; o["Serialization"] = n
        n = Retiring() ; r.run(n) ; o["Retiring"] = n

        # parents

        o["IFetch_Latency"].parent = o["Frontend_Bound"]
        o["ICache_Misses"].parent = o["IFetch_Latency"]
        o["ITLB_Misses"].parent = o["IFetch_Latency"]
        o["Branch_Detect"].parent = o["IFetch_Latency"]
        o["Branch_Resteer"].parent = o["IFetch_Latency"]
        o["IFetch_Bandwidth"].parent = o["Frontend_Bound"]
        o["Cisc"].parent = o["IFetch_Bandwidth"]
        o["Decode"].parent = o["IFetch_Bandwidth"]
        o["Predecode"].parent = o["IFetch_Bandwidth"]
        o["Other_FB"].parent = o["IFetch_Bandwidth"]
        o["Branch_Mispredicts"].parent = o["Bad_Speculation"]
        o["Machine_Clears"].parent = o["Bad_Speculation"]
        o["Nuke"].parent = o["Machine_Clears"]
        o["Fast_Nuke"].parent = o["Machine_Clears"]
        o["Core_Bound"].parent = o["Backend_Bound"]
        o["Allocation_Restriction"].parent = o["Core_Bound"]
        o["Resource_Bound"].parent = o["Backend_Bound"]
        o["Mem_Scheduler"].parent = o["Resource_Bound"]
        o["Non_Mem_Scheduler"].parent = o["Resource_Bound"]
        o["Register"].parent = o["Resource_Bound"]
        o["Reorder_Buffer"].parent = o["Resource_Bound"]
        o["Serialization"].parent = o["Resource_Bound"]

        # user visible metrics

        n = Metric_PCT_Mem_Exec_Bound_Cycles() ; r.metric(n) ; o["PCT_Mem_Exec_Bound_Cycles"] = n
        n = Metric_PCT_Load_Miss_Bound_Cycles() ; r.metric(n) ; o["PCT_Load_Miss_Bound_Cycles"] = n
        n = Metric_PCT_DTLB_Miss_Bound_Cycles() ; r.metric(n) ; o["PCT_DTLB_Miss_Bound_Cycles"] = n
        n = Metric_PCT_IFetch_Miss_Bound_Cycles() ; r.metric(n) ; o["PCT_IFetch_Miss_Bound_Cycles"] = n
        n = Metric_IPC() ; r.metric(n) ; o["IPC"] = n
        n = Metric_CPI() ; r.metric(n) ; o["CPI"] = n
        n = Metric_UPI() ; r.metric(n) ; o["UPI"] = n
        n = Metric_FLOPc() ; r.metric(n) ; o["FLOPc"] = n
        n = Metric_IpBranch() ; r.metric(n) ; o["IpBranch"] = n
        n = Metric_IpCall() ; r.metric(n) ; o["IpCall"] = n
        n = Metric_IpFarBranch() ; r.metric(n) ; o["IpFarBranch"] = n
        n = Metric_IpMispredict() ; r.metric(n) ; o["IpMispredict"] = n
        n = Metric_IpMisp_Cond_Ntaken() ; r.metric(n) ; o["IpMisp_Cond_Ntaken"] = n
        n = Metric_IpMisp_Cond_Taken() ; r.metric(n) ; o["IpMisp_Cond_Taken"] = n
        n = Metric_IpMisp_Ret() ; r.metric(n) ; o["IpMisp_Ret"] = n
        n = Metric_IpMisp_Indirect() ; r.metric(n) ; o["IpMisp_Indirect"] = n
        n = Metric_IpFLOP() ; r.metric(n) ; o["IpFLOP"] = n
        n = Metric_IpFPArith() ; r.metric(n) ; o["IpFPArith"] = n
        n = Metric_IpFPArith_Scalar_SP() ; r.metric(n) ; o["IpFPArith_Scalar_SP"] = n
        n = Metric_IpFPArith_Scalar_DP() ; r.metric(n) ; o["IpFPArith_Scalar_DP"] = n
        n = Metric_IpFPArith_AVX128() ; r.metric(n) ; o["IpFPArith_AVX128"] = n
        n = Metric_IpFPArith_AVX256() ; r.metric(n) ; o["IpFPArith_AVX256"] = n
        n = Metric_Microcode_Uop_Ratio() ; r.metric(n) ; o["Microcode_Uop_Ratio"] = n
        n = Metric_FPDiv_Uop_Ratio() ; r.metric(n) ; o["FPDiv_Uop_Ratio"] = n
        n = Metric_IDiv_Uop_Ratio() ; r.metric(n) ; o["IDiv_Uop_Ratio"] = n
        n = Metric_X87_Uop_Ratio() ; r.metric(n) ; o["X87_Uop_Ratio"] = n
        n = Metric_IpLoad() ; r.metric(n) ; o["IpLoad"] = n
        n = Metric_IpStore() ; r.metric(n) ; o["IpStore"] = n
        n = Metric_MemLoad_Ratio() ; r.metric(n) ; o["MemLoad_Ratio"] = n
        n = Metric_Load_Splits_Ratio() ; r.metric(n) ; o["Load_Splits_Ratio"] = n
        n = Metric_Load_Locks_Ratio() ; r.metric(n) ; o["Load_Locks_Ratio"] = n
        n = Metric_PCT_Loads_with_StoreFwdBlk() ; r.metric(n) ; o["PCT_Loads_with_StoreFwdBlk"] = n
        n = Metric_PCT_Loads_with_AdressAliasing() ; r.metric(n) ; o["PCT_Loads_with_AdressAliasing"] = n
        n = Metric_PCT_LoadHead_with_STLBHit() ; r.metric(n) ; o["PCT_LoadHead_with_STLBHit"] = n
        n = Metric_PCT_LoadHead_with_Pagewalk() ; r.metric(n) ; o["PCT_LoadHead_with_Pagewalk"] = n
        n = Metric_PCT_LoadHead_with_StoreFwding() ; r.metric(n) ; o["PCT_LoadHead_with_StoreFwding"] = n
        n = Metric_PCT_LoadHead_with_OtherPipelineBlks() ; r.metric(n) ; o["PCT_LoadHead_with_OtherPipelineBlks"] = n
        n = Metric_PCT_LoadHead_with_L1miss() ; r.metric(n) ; o["PCT_LoadHead_with_L1miss"] = n
        n = Metric_Store_Bound() ; r.metric(n) ; o["Store_Bound"] = n
        n = Metric_Load_Bound() ; r.metric(n) ; o["Load_Bound"] = n
        n = Metric_L1_Bound() ; r.metric(n) ; o["L1_Bound"] = n
        n = Metric_PCT_LoadMissBound_with_L2Hit() ; r.metric(n) ; o["PCT_LoadMissBound_with_L2Hit"] = n
        n = Metric_PCT_LoadMissBound_with_L2Miss() ; r.metric(n) ; o["PCT_LoadMissBound_with_L2Miss"] = n
        n = Metric_PCT_IfetchMissBound_with_L2Hit() ; r.metric(n) ; o["PCT_IfetchMissBound_with_L2Hit"] = n
        n = Metric_PCT_IfetchMissBound_with_L2Miss() ; r.metric(n) ; o["PCT_IfetchMissBound_with_L2Miss"] = n
        n = Metric_Branch_Mispredict_Ratio() ; r.metric(n) ; o["Branch_Mispredict_Ratio"] = n
        n = Metric_Branch_Mispredict_to_Unknown_Branch_Ratio() ; r.metric(n) ; o["Branch_Mispredict_to_Unknown_Branch_Ratio"] = n
        n = Metric_Machine_Clears_Disamb_PKI() ; r.metric(n) ; o["Machine_Clears_Disamb_PKI"] = n
        n = Metric_Machine_Clears_FP_Assist_PKI() ; r.metric(n) ; o["Machine_Clears_FP_Assist_PKI"] = n
        n = Metric_Machine_Clears_Monuke_PKI() ; r.metric(n) ; o["Machine_Clears_Monuke_PKI"] = n
        n = Metric_Machine_Clears_Page_Fault_PKI() ; r.metric(n) ; o["Machine_Clears_Page_Fault_PKI"] = n
        n = Metric_Machine_Clears_MRN_PKI() ; r.metric(n) ; o["Machine_Clears_MRN_PKI"] = n
        n = Metric_PCT_Store_Buffer_Stall_Cycles() ; r.metric(n) ; o["PCT_Store_Buffer_Stall_Cycles"] = n
        n = Metric_PCT_Load_Buffer_Stall_Cycles() ; r.metric(n) ; o["PCT_Load_Buffer_Stall_Cycles"] = n
        n = Metric_PCT_Mem_RSV_Stall_Cycles() ; r.metric(n) ; o["PCT_Mem_RSV_Stall_Cycles"] = n
        n = Metric_PCT_Tpause_Cycles() ; r.metric(n) ; o["PCT_Tpause_Cycles"] = n
        n = Metric_Turbo_Utilization() ; r.metric(n) ; o["Turbo_Utilization"] = n
        n = Metric_Kernel_Utilization() ; r.metric(n) ; o["Kernel_Utilization"] = n
        n = Metric_CPU_Utilization() ; r.metric(n) ; o["CPU_Utilization"] = n
        n = Metric_GFLOPs() ; r.metric(n) ; o["GFLOPs"] = n

        # references between groups

        o["Resource_Bound"].Core_Bound = o["Core_Bound"]

        # siblings cross-tree




# Handle warnings and errors
# Separate module to avoid circular imports
from __future__ import print_function
import sys
import fnmatch

quiet = False
debug = None

def debug_msg(x, y):
    if debug and any(map(lambda p: fnmatch.fnmatch(x, p), debug.split(","))):
        print("debug:", x + ": " + str(y), file=sys.stderr)

def warning(x):
    if not quiet:
        print("WARNING:", x, file=sys.stderr)



# utility functions

import bisect

def find_le(f, key):
    pos = bisect.bisect_left(f, (key,))
    if pos < len(f) and f[pos][0] == key:
        return f[pos]
    if pos == 0:
        return None
    return f[pos - 1]



#
# TSX metrics
#

# XXX force all these into a single group
# XXX: force % in caller

import os

def TXCycles(EV, level):
    return EV("cpu/cycles-t/", level) / EV("cycles", level)

class TransactionalCycles:
    name = "Transactional cycles"
    desc = """
Percent cycles spent in a transaction. When low or zero either the program
does not use locks (or other transactions), or the locks are not enabled with lock elision."""
    subplot = "TSX"
    unit = "%"
    sample = ["mem_uops_retired.lock_loads"]
    server = True
    def compute(self, EV):
        try:
            self.val = TXCycles(EV, 1) * 100.
            self.thresh = (self.val >= 0.01)
        except ZeroDivisionError:
            self.val = 0
            self.thresh = False

class AbortedCycles:
    name = "Aborted cycles"
    desc = """
Percent cycles wasted in transaction aborts. When a significant part of the transactional cycles
start sampling for abort causes."""
    subplot = "TSX"
    unit = "%"
    sample = ["cpu/tx-abort/pp", "cpu/hle-abort/pp"]
    server = True
    def compute(self, EV):
        try:
            self.val = ((EV("cpu/cycles-t/", 1) - EV("cpu/cycles-ct/", 1)) / EV("cycles", 1)) * 100.
            self.thresh = (self.val >= 0.01)
        except ZeroDivisionError:
            self.val = 0
            self.thresh = False

class AverageRTM:
    name = "Average RTM transaction length"
    desc = """
Average RTM transaction length. Assumes most transactions are RTM.
When low consider increasing the size of the critical sections to lower overhead."""
    subplot = "TSX Latencies"
    unit = "cycles"
    server = True
    def compute(self, EV):
        try:
            self.val = EV("cpu/cycles-t/", 1) / EV("RTM_RETIRED.START", 1)
            self.thresh = TXCycles(EV, 1) >= 0.01 and self.val > 0
        except ZeroDivisionError:
            self.val = 0
            self.thresh = False

class AverageHLE:
    name = "Average HLE transaction length"
    desc = """
Average HLE transaction length. Assumes most transactions are HLE.
When low consider increasing the size of the critical sections to lower overhead."""
    subplot = "TSX Latencies"
    unit = "cycles"
    def compute(self, EV):
        try:
            self.val = EV("cpu/cycles-t/", 1) / EV("HLE_RETIRED.START", 1)
            self.thresh = TXCycles(EV, 1) >= 0.01 and self.val > 0
        except ZeroDivisionError:
            self.val = 0
            self.thresh = False

class Setup:
    def __init__(self, r):
        # XXX allow override
        if os.path.exists("/sys/bus/event_source/devices/cpu/events/cycles-t"):
            r.force_metric(TransactionalCycles())
            r.force_metric(AbortedCycles())
            r.force_metric(AverageRTM())
            #r.force_metric(AverageHLE())



#!/usr/bin/env python3
# print all events in a eventmap
from __future__ import print_function
import sys
import ocperf
emap = ocperf.find_emap()
if not emap:
    sys.exit("Unknown CPU or cannot find event table")
for j in sorted(emap.events):
    print(j)



# generate levels for events from the model
# utility module for other tools
l1 = set(("Frontend_Bound", "Backend_Bound", "Retiring", "Bad_Speculation"))

def get_level(name):
    is_node = name in l1 or "." in name
    level = name.count(".") + 1
    if is_node:
        return level
    return 0

def is_metric(name):
    return get_level(name) == 0

def level_name(name):
    if name.count(".") > 0:
        f = name.split(".")[:-1]
        n = ".".join(f)
    elif is_metric(name):
        return "CPU-METRIC" # XXX split
    else:
        n = "TopLevel"
    n = n.replace(" ", "_")
    return n



#
# perf power metrics for toplev
#

import os

class EnergyPackage:
    name = "Package Energy"
    desc = """
Package Energy over measurement period in Joules"""
    unit = "Joules"
    nogroup = True
    subplot = "Power"
    domain = "Package"
    def compute(self, EV):
        self.val = EV("power/energy-pkg/", 1)
        self.thresh = self.val > 0

class EnergyCores:
    name = "Cores Energy"
    desc = """
Cores Energy over measurement period in Joules"""
    unit = "Joules"
    nogroup = True
    subplot = "Power"
    domain = "Package"
    def compute(self, EV):
        self.val = EV("power/energy-cores/", 1)
        self.thresh = self.val > 0

class EnergyRAM:
    name = "RAM Energy"
    desc = """
RAM Energy over measurement period in Joules"""
    unit = "Joules"
    nogroup = True
    subplot = "Power"
    domain = "Package"
    def compute(self, EV):
        self.val = EV("power/energy-ram/", 1)
        self.thresh = self.val > 0

class EnergyGPU:
    name = "GPU Energy"
    desc = """
GPU Energy over measurement period in Joules"""
    unit = "Joules"
    nogroup = True
    subplot = "Power"
    domain = "Package"
    def compute(self, EV):
        self.val = EV("power/energy-gpu/", 1)
        self.thresh = self.val > 1

class Setup:
    def __init__(self, r):
        if os.path.exists("/sys/bus/event_source/devices/power/events/energy-cores"):
            r.force_metric(EnergyCores())
        r.force_metric(EnergyPackage())
        if os.path.exists("/sys/bus/event_source/devices/power/events/energy-ram"):
            r.force_metric(EnergyRAM())
        if os.path.exists("/sys/bus/event_source/devices/power/events/energy-gpu"):
            r.force_metric(EnergyGPU())



#!/usr/bin/env python3
# Do basic python sanity check of translation output
import sys
sys.path.append(".")
import t
l = []
m = []

def pev(e):
    print("\t",e)
    return 1

class R:
    def run(self, p):
        #print p
        l.append(p)
    def metric(self, p):
        m.append(p)

t.Setup(R())
for p in l:
    p.thresh = True
for p in l:
    print(p.name)
    p.compute(lambda e, level: pev(e))
    if p.sample:
        print("    Sample:", " ".join(p.sample))
    if p.sibling:
        print("    Siblings:", " ".join([o.name for o in p.sibling]))

for p in m:
    print(p.name)
    p.compute(lambda e, level: pev(e))




# Copyright (c) 2020, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
from __future__ import print_function
import sys
import subprocess
import os
import argparse
if sys.version_info.major == 3:
    from typing import Set # noqa

if sys.version_info.major == 3:
    popentext = dict(universal_newlines=True)
else:
    popentext = {}

def popen_stdout(cmd):
    return subprocess.Popen(cmd, stdout=subprocess.PIPE, **popentext) # type: ignore

def popen_stdinout(cmd, f):
    return subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=f, **popentext) # type: ignore

def flex_open_r(fn):
    if fn.endswith(".xz"):
        xz = popen_stdout(["xz", "-d", "--stdout", fn])
        return xz.stdout
    if fn.endswith(".gz"):
        gzip = popen_stdout(["gzip", "-d", "-c", fn])
        return gzip.stdout
    if fn.endswith(".zst"):
        return popen_stdout(["zstd", "-d", "--stdout", fn]) .stdout
    return open(fn, 'r')

def flex_open_w(fn):
    f = open(fn, "w")
    if fn.endswith(".xz"):
        xz = popen_stdinout(["xz", "-z", "--stdout"], f)
        return xz.stdin
    if fn.endswith(".gz"):
        gzip = popen_stdinout(["gzip", "-c"], f)
        return gzip.stdin
    if fn.endswith(".zst"):
        return popen_stdinout(["zstd", "--stdout"], f).stdin
    return f

tl_tester = os.getenv("TL_TESTER")
test_mode = tl_tester and tl_tester != "0"

args = argparse.Namespace()

def set_args(a):
    global args
    args = a

def warn_no_assert(msg):
    if not args.quiet:
        print("warning: " + msg, file=sys.stderr)

def warn_test(msg):
    if test_mode:
        warn_no_assert(msg)

def warn(msg):
    warn_no_assert(msg)
    if test_mode:
        assert 0, msg

warned = set() # type: Set[str]

def warn_once_no_assert(msg):
    if msg not in warned and not args.quiet:
        print("warning: " + msg, file=sys.stderr)
        warned.add(msg)

def warn_once(msg):
    warn_once_no_assert(msg)
    if test_mode:
        assert 0, msg

def print_once(msg):
    if msg not in warned and not args.quiet:
        print(msg)
        warned.add(msg)

def inform(msg):
    if not args.quiet:
        print(msg)

def debug_print(x):
    if args.debug:
        print(x, file=sys.stderr)

def obj_debug_print(obj, x):
    if args.debug or (args.dfilter and obj.name in args.dfilter):
        print(x, file=sys.stderr)

def test_debug_print(x):
    if args.debug or test_mode:
        print(x, file=sys.stderr)



#!/usr/bin/env python3
# convert perf stat -Ixxx -x, / toplev -Ixxx -x, output to normalized output
# this version buffers all data in memory, so it can use a lot of memory.
# t1,ev1,num1
# t1,ev2,num1
# t2,ev1,num3
# ->
# timestamp,ev1,ev2
# t1,num1,num2
# t2,num3,,
# when the input has CPU generate separate lines for each CPU (may need post filtering)
from __future__ import print_function
import sys
import csv
import argparse
import collections
import csv_formats

ap = argparse.ArgumentParser(description=
'Normalize CSV data from perf or toplev. All values are printed on a single line.')
ap.add_argument('inputfile', type=argparse.FileType('r'), default=sys.stdin, nargs='?')
ap.add_argument('--output', '-o', type=argparse.FileType('w'), default=sys.stdout, nargs='?')
ap.add_argument('--cpu', nargs='?', help='Only output for this cpu')
ap.add_argument('--na', nargs='?', help='Value to use if data is not available', default="")
ap.add_argument('--error-exit', action='store_true', help='Force error exit on parse error')
ap.add_argument('--normalize-cpu', action='store_true', help='Normalize CPUs into unique columns too')
args = ap.parse_args()

printed_header = False
timestamp = None

events = collections.OrderedDict()
out = []
times = []
cpus = []
rc = csv.reader(args.inputfile)
res = []
writer = csv.writer(args.output, lineterminator='\n')
lastcpu = None
cpu = None
lineno = 1
for row in rc:
    if len(row) > 0 and (row[0] == "Timestamp" or row[0].startswith("#")):
        lineno += 1
        continue
    r = csv_formats.parse_csv_row(row, error_exit=args.error_exit)
    if r is None:
        print("at line %d" % lineno, file=sys.stderr)
        lineno += 1
        continue
    ts, cpu, ev, val = r.ts, r.cpu, r.ev, r.val

    if ts != timestamp or (cpu != lastcpu and not args.normalize_cpu):
        if timestamp:
            if args.cpu and cpu != args.cpu:
                continue
            # delay in case we didn't see all headers
            # only need to do that for toplev, directly output for perf?
            # could limit buffering to save memory?
            out.append(res)
            times.append(timestamp)
            cpus.append(cpu)
            res = []
        timestamp = ts
        lastcpu = cpu

    if cpu is not None and args.normalize_cpu:
        ev = cpu + " " + ev

    # use a list for row storage to keep memory requirements down
    if ev not in events:
        events[ev] = len(res)
    ind = events[ev]
    if ind >= len(res):
        res += [None] * ((ind + 1) - len(res))
    res[ind] = val
    lineno += 1
if res and not (args.cpu and cpu != args.cpu):
    out.append(res)
    times.append(timestamp)
    cpus.append(cpu)

def resolve(row, ind):
    if ind >= len(row):
        return args.na
    v = row[ind]
    if v is None:
        return args.na
    return v

def cpulist():
    if args.normalize_cpu:
        return []
    if cpu is not None:
        return ["CPU"]
    return []

keys = events.keys()
writer.writerow(["Timestamp"] + cpulist() + list(keys))
for row, ts, cpunum in zip(out, times, cpus):
    writer.writerow([ts] +
                ([cpunum] if (cpu is not None and not args.normalize_cpu) else []) +
                ([resolve(row, events[x]) for x in keys]))



#!/usr/bin/env python3
# serve toplev csv file as http using dygraph
# toplev.py -I100 -o x.csv -v -x, ...
# tl-serve.py x.csv [host [port]]
from __future__ import print_function
import string
import argparse
import csv
import re
import os
import signal
import sys
try:
    import BaseHTTPServer
except ImportError:
    import http.server as BaseHTTPServer
import tldata

ap = argparse.ArgumentParser(usage="Serve toplev csv file as http or generate in directory")
ap.add_argument('csvfile', help='toplev csv file to serve')
ap.add_argument('host', nargs='?', default="localhost", help='Hostname to bind to (default localhost)')
ap.add_argument('port', nargs='?', default="9001", type=int, help='Port to bind to (default 9001)')
ap.add_argument('--verbose', '-v', action='store_true', help='Display all metrics, even if below threshold')
ap.add_argument('--gen', help='Generate HTML files in specified directory')
ap.add_argument('--title', help='Title for output')
args = ap.parse_args()

T = string.Template

data = tldata.TLData(args.csvfile, args.verbose)
data.update()

def jsname(n):
    return n.replace(".", "_").replace("-", "_")

def comma_strings(l):
    return ",".join(['"%s"' % (x) for x in l])

def gen_html_header():
    graph = """<html><head><title>Toplev</title>
<link rel="shortcut icon" href="toplev.ico" />
<script type="text/javascript" src="dygraph-combined.js"></script>
</head>
<body>
<script type="text/javascript">

var cpus = ["""

    graph += comma_strings(sorted(data.cpus)) + "]\n"
    graph += "var nodes = [" + comma_strings(tldata.level_order(data)) + "]"

    graph += """
var graphs = []
var goptions = []
var num_graphs = 0
var block_redraw = false

function enable(el) {
    for (i = 0; i < cpus.length; i++) {
        p = document.getElementById("d_" + cpus[i] + "_" + el.name)
        p.style.display = el.checked ? 'block' : 'none';
    }
}

function change_all(flag) {
    all_displays = document.getElementsByClassName("disp")
    for (i = 0; i < all_displays.length; i++) {
        p = all_displays[i];
        p.style.display = flag ? 'block' : 'none';
    }
    togs = document.getElementsByClassName("toggles")
    for (i = 0; i < togs.length; i++) {
        p = togs[i];
        p.checked = flag;
    }
}

var timer

function toggle_refresh(el) {
    p = document.getElementById("refresh_rate")
    if (timer) {
        clearInterval(timer)
        timer = null
    }
    if (el.checked) {
        timer = setInterval(function () {
                    for (i = 0; i < num_graphs; i++) {
                        graphs[i].updateOptions(goptions[i])
                    }
                }, Number(p.value))
    }
}

function add_node(cpu, nd) {
    p = document.getElementById("d_" + cpus[cpu] + "_" + nodes[nd])
    p.parentNode.appendChild(p)
}

function toggle_interleave(el) {
    if (!el.checked) {
        for (cpu = 0; cpu < cpus.length; cpu++) {
            for (nd = 0; nd < nodes.length; nd++) {
                add_node(cpu, nd)
            }
        }
    } else {
        for (nd = 0; nd < nodes.length; nd++) {
            for (cpu = 0; cpu < cpus.length; cpu++) {
                add_node(cpu, nd)
            }
        }
    }
}

function draw_graph(me, initial) {
    if (block_redraw || initial)
        return;
    block_redraw = true
    xrange = me.xAxisRange()
    for (i = 0; i < num_graphs; i++) {
        if (graphs[i] != me) {
            graphs[i].updateOptions({
                dateWindow: xrange,
            })
        }
    }
    block_redraw = false
}

function hilight_help(e, x, pts, row, help) {
    p = document.getElementById("help")
    h = ""
    for (i = 0; i < pts.length; i++) {
        n = pts[i].name
        if (n in help && help[n] != "") {
            h += "<b>" + n + "</b>: " + help[n] + " <br /> "
        } else {
            // location.reload(); // XXX
        }
    }
    p.innerHTML = h
}

function unhilight_help(e, x, pts, row) {
    p = document.getElementById("help")
    p.innerHTML = ""
}

</script>
"""
    if args.title:
        graph += T("<h1>$title</h1>\n").substitute({"title": args.title})
    graph += """
<div><p>
<b>Display:</b>
"""
    lev = tldata.level_order(data)
    for num, name in enumerate(lev):
        graph += T("""\
<input id="$id" class="toggles" type=checkbox name="$name" onClick="enable(this)" checked />
<label for="$id">$name</label>
        """).substitute({"id": num, "name": name})
    graph += """
<input id="all" type=checkbox name="dall" onClick="change_all(this.checked)" checked />
<label for="all">Toggle all</label>
<input id="enable_refresh" type=checkbox onClick="toggle_refresh(this)" />
<label for="enable_refresh">Auto-refresh</label>
<input id="refresh_rate" type="text" size=4 value="1000" name="refresh"  />
<label for="refresh_rate">Refresh rate (ms)</label>
<input id="interleave" type=checkbox onClick="toggle_interleave(this)" />
<label for="interleave">Interleave CPUs</label>
</p></div>

Drag to zoom. Double click to zoom out again<br />

<div id="help" style="position:fixed; right:0; width:300px; font-size: 11"> </div>
"""

    for j in lev:
        graph += T("""
<script type="text/javascript">
help_$name = {
""").substitute({"name": jsname(j)})
        for i in data.levels[j]:
            if i not in data.helptxt:
                #print(i,"not found in",data.helptxt.keys())
                continue
            graph += T("""
        "$name": "$help",
        """).substitute({"name": get_postfix(i), "help": data.helptxt[i]})
        graph += """
}
</script>
"""
    return graph

def gen_html_cpu(cpu):
    lev = tldata.level_order(data)
    graph = ""
    for name in lev:
        opts = {
            "title": name + " " + cpu,
            "width": 1000,
            "height": 180,
            #"xlabel": "time",
        }

        if name in data.metrics:
            unit = None
            if name in data.units: # XXX handle CPU-METRIC
                unit = data.units[name]
            if unit:
                opts["ylabel"] = unit
        else:
            opts["stackedGraph"] = 1
            opts["stackedGraphNaNFill"] = "none"
            opts["ylabel"] = "% CPU time"
            unit = '%'
        if unit == '%':
            opts["valueRange"] = [-5, 110]
        graph += T("""

<div id="d_${cpu}_$name" class="disp"></div>

<script type="text/javascript">
    i = num_graphs++
    goptions[i] = $opts
    goptions[i].highlightCallback = function(e, x, pts, row) {
        hilight_help(e, x, pts, row, help_$jname)
    }
    goptions[i].unhighlightCallback = unhilight_help
    goptions[i].drawCallback = draw_graph
    graphs[i] = new Dygraph(document.getElementById("d_${cpu}_$name"), "$cpu.$file.csv", goptions[i])
    goptions[i]["file"] = "$cpu.$file.csv"
</script>
                """).substitute({"name": name, "jname": jsname(name), "file": name, "cpu": cpu, "opts": opts})
    return graph

def gen_html():
    graph = gen_html_header()
    for cpu in sorted(data.cpus):
        graph += gen_html_cpu(cpu)
    graph += """
</body>
</html>"""
    return graph

def get_postfix(s):
    m = re.match(r'.*\.(.*)', s)
    if m:
        return m.group(1)
    return s

def gencsv(wfile, l, cpu):
    hdr = sorted(data.levels[l])
    wr = csv.writer(wfile, lineterminator='\n')
    wr.writerow(["Timestamp"] + map(get_postfix, hdr))
    for val, ts in zip(data.vals, data.times):
        wr.writerow([ts] + [val[(x, cpu)] if (x, cpu) in val else "" for x in hdr])

class TLHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    def header(self, typ):
        self.send_response(200)
        self.send_header('Content-Type', typ)
        self.end_headers()

    def bad(self):
        self.send_response(401)
        self.send_header('Content-Type', 'text/html')
        self.end_headers()
        self.wfile.write("%s not found" % (self.path))

    def serve_file(self, fn, mime):
        with open(fn, "r") as f:
            self.header(mime)
            self.wfile.write(f.read().encode('utf-8'))

    def do_GET(self):
        if self.path == "/":
            self.header("text/html")
            self.wfile.write(gen_html().encode())
        elif self.path == "/dygraph-combined.js":
            self.serve_file("dygraph-combined.js", "text/javascript")
        elif self.path == "/toplev.ico":
            self.serve_file("toplev.ico", "image/x-icon")
        elif self.path.endswith(".csv"):
            data.update()
            m = re.match(r"/(cpu|C\d+|S\d+-C\d+|C\d+-T\d+)\.(.*?)\.csv", self.path)
            if not m:
                self.bad()
                return
            cpu = m.group(1)
            l = m.group(2)
            if l not in data.levels:
                self.bad()
                return
            self.header("text/csv")
            gencsv(self.wfile, l, cpu)
        else:
            self.bad()

def copyfile(a, b):
    with open(a, "rb") as af:
        with open(b, "wb") as bf:
            bf.write(af.read())

def term(signal, frame):
    print("sigterm")
    sys.exit(0)

if args.gen:
    if not os.path.isdir(args.gen):
        os.makedirs(args.gen)
    genfn = os.path.join
    with open(genfn(args.gen, "index.html"), 'wb') as f:
        f.write(gen_html().encode())
    copyfile('dygraph-combined.js', genfn(args.gen, 'dygraph-combined.js'))
    copyfile('toplev.ico', genfn(args.gen, 'favicon.ico'))
    for cpu in data.cpus:
        for l in data.levels:
            with open(genfn(args.gen, cpu + "." + l + ".csv"), 'w') as f:
                gencsv(f, l, cpu)
    print("Please browse", args.gen, "through a web server, not through file:")
else:
    signal.signal(signal.SIGTERM, term)

    httpd = BaseHTTPServer.HTTPServer((args.host, args.port), TLHandler)

    print("serving at",args.host,"port",args.port,"until Ctrl-C")
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        httpd.socket.close()



# Copyright (c) 2012-2025, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# Maintain error data on perf measurements
from __future__ import print_function
import math
from collections import namedtuple
from tl_io import warn, warn_test, inform

ValStat = namedtuple('ValStat', ['stddev', 'multiplex'])

def geoadd(l):
    return math.sqrt(sum([x**2 for x in l]))

# use geomean of stddevs and minimum of multiplex ratios for combining
# XXX better way to combine multiplex ratios?
def combine_valstat(l):
    if not l:
        return []
    return ValStat(geoadd([x.stddev for x in l]), min([x.multiplex for x in l]))

class ComputeStat:
    """Maintain statistics on measurement data."""
    def __init__(self, quiet):
        self.referenced = set()
        self.already_warned = set()
        self.errcount = 0
        self.errors = set()
        self.prev_errors = set()
        self.mismeasured = set()
        self.prev_mismeasured = set()
        self.quiet = quiet

    def referenced_check(self, res, evnum):
        referenced = self.referenced
        referenced = referenced - self.already_warned
        if not referenced:
            return
        self.already_warned |= referenced

        # sanity check: did we reference all results?
        if len(res.keys()) > 0:
            r = res[list(res.keys())[0]]
            if len(r) != len(evnum):
                warn("results len %d does not match event len %d" % (len(r), len(evnum)))
                return
            if len(referenced) != len(r):
                dummies = {i for i, d in enumerate(evnum) if d == "dummy"}
                notr = set(range(len(r))) - referenced - dummies
                if notr:
                    warn_test("%d results not referenced: " % (len(notr)) +
                          " ".join(["%d" % x for x in sorted(notr)]))

    def compute_errors(self):
        if self.errcount > 0 and self.errors != self.prev_errors:
            inform(("%d nodes had zero counts: " % (self.errcount)) +
                           " ".join(sorted(self.errors)))
            self.errcount = 0
            self.prev_errors = self.errors
            self.errors = set()
        if self.mismeasured and self.mismeasured > self.prev_mismeasured:
            inform("Mismeasured (out of bound values):" +
                    " ".join(sorted(self.mismeasured)))
            self.prev_mismeasured = self.mismeasured



nominal_freq = 1.0

class Frequency:
    name = "Frequency"
    desc = " Frequency ratio"
    subplot = "Frequency"
    domain = "CoreMetric"

    def compute(self, EV):
        try:
            self.val = (EV("cycles", 1) / EV("CPU_CLK_UNHALTED.REF_TSC", 1)) * nominal_freq
        except ZeroDivisionError:
            self.val = 0

class SetupCPU:
    def __init__(self, r, cpu):
        global nominal_freq
        if cpu.freq:
            nominal_freq = cpu.freq
        r.force_metric(Frequency())



#!/usr/bin/env python3
# merge multiple --perf-output files. requires header
from __future__ import print_function
import csv
import argparse
from collections import OrderedDict, Counter
import sys

ap = argparse.ArgumentParser()
ap.add_argument('csvfiles', nargs='+', type=argparse.FileType('r'))
args = ap.parse_args()

def genkey(c, hdr, count):
    k = [count]
    if 'Timestamp' in hdr:
        k.append(c[hdr['Timestamp']])
    if 'Location' in hdr:
        k.append(c[hdr['Location']])
    k.append(c[hdr['Event']])
    return tuple(k)

d = OrderedDict()
hdr = None
hdrl = None
prev = Counter()
for fh in args.csvfiles:
    csvf = csv.reader(fh, delimiter=';')
    for c in csvf:
        if hdr is None:
            hdrl = c
            hdr = dict([(y,x) for x,y in enumerate(c)])
            continue
        if c[0] in ("Timestamp", "Location", "Value"):
            continue
        pkey = (fh, c[hdr['Timestamp']] if 'Timestamp' in hdr else None, c[hdr['Event']])
        prev[pkey] += 1
        key = genkey(c, hdr, prev[pkey])
        try:
            if key in d:
                o = d[key]
                o[hdr['Run-Time']] += float(c[hdr['Run-Time']])
                o[hdr['Enabled']] = (float(o[hdr['Enabled']]) + o[hdr['Enabled']]) / 2
                o[hdr['Value']] += float(c[hdr['Value']])
            else:
                d[key] = c
                o = d[key]
                o[hdr['Value']] = float(c[hdr['Value']])
                o[hdr['Enabled']] = float(c[hdr['Enabled']])
                o[hdr['Run-Time']] = float(c[hdr['Run-Time']])
        except ValueError as e:
            print("cannot parse", c, e, file=sys.stderr)

csvf = csv.writer(sys.stdout, delimiter=';')
csvf.writerow(hdrl)
for j in d.values():
    csvf.writerow(j)



#
# basic linux software metrics
#
# for most of these it would be much nicer to measure durations, but
# perf stat doesn't support that
#
from __future__ import print_function
import os, sys

class CS:
    name = "Context switches"
    desc = " Number of context switches between threads or processes on a CPU."
    nogroup = True
    subplot = "Scheduler"
    def compute(self, EV):
        self.val = EV("cs", 1)
        self.thresh = self.val > 0

class MinorFaults:
    name = "Minor faults"
    desc = " Page faults not leading to disk IO, such as allocation of memory."
    nogroup = True
    subplot = "Faults"
    def compute(self, EV):
        self.val = EV("minor-faults", 1)
        self.thresh = self.val > 0

class MajorFaults:
    name = "Major faults"
    desc = """
Page faults leading to disk IO, such as loading executable text or do mmap'ed IO."""
    nogroup = True
    subplot = "Faults"
    def compute(self, EV):
        self.val = EV("major-faults", 1)
        self.thresh = self.val > 0

class Migrations:
    name = "Migrations"
    desc = " Number of thread/process migrations to another CPU."
    nogroup = True
    subplot = "Scheduler"
    def compute(self, EV):
        self.val = EV("migrations", 1)
        self.thresh = self.val > 0

# The events below require trace points, so typically root.

class BadEV(Exception):
    pass

def C(ev):
    if not os.path.exists("/sys/kernel/debug/tracing/events" + ev.replace(":", "/")):
        raise BadEV()
    return ev

class Syscalls:
    name = "Syscalls"
    desc = " Number of syscalls, not including vsyscalls such as gettimeofday."
    nogroup = True
    subplot = "OS metrics"
    def compute(self, EV):
        try:
            self.val = EV(C("raw_syscalls:sys_enter"), 1)
        except BadEV:
            self.val = 0
        self.thresh = self.val > 0

class Interrupts:
    name = "Interrupts"
    desc = """
Number of interrupts, including NMIs, excluding exceptions. These are interrupts caused by hardware,
typically to indicate IO. This includes performance counter sampling interrupts."""
    nogroup = True
    subplot = "Interrupts"
    # can overcount with shared vectors
    def compute(self, EV):
        try:
            self.val = EV(C("irq:irq_handler_entry"), 1) + EV(C("nmi:nmi_handler"), 1)
        except BadEV:
            self.val = 0
        self.thresh = self.val > 0

# XXX on older kernels will not count TLB flushes, when they still had an
# own vector
class IPIs:
    name = "IPIs"
    desc = """
Number of inter-processor-interrupts (IPIs). These are caused by software, for example to flush TLBs,
finish IOs on the originating CPU, flush per CPU software caches (such as slab allocator caches)
or force reschedules."""
    nogroup = True
    subplot = "Interrupts"
    # can overcount with shared vectors
    def compute(self, EV):
        try:
            self.val = (EV(C("irq_vectors:call_function_entry"), 1) +
                        EV(C("irq_vectors:call_function_single_entry"), 1) +
                        EV(C("irq_vectors:reschedule_entry"), 1))
        except BadEV:
            self.val = 0
        self.thresh = self.val > 0

class Workqueues:
    name = "Workqueues"
    desc = " Work queue item executions. These are tasks executed by the kernel in the background."
    nogroup = True
    subplot = "Interrupts"
    def compute(self, EV):
        try:
            self.val = EV(C("workqueue:workqueue_execute_start"), 1)
        except BadEV:
            self.val = 0
        self.thresh = self.val > 0

class BlockIOs:
    name = "BlockIOs"
    desc = " Block IOs issued. This counts the number of block IO requests inserted into a queue."
    nogroup = True
    subplot = "IO"
    def compute(self, EV):
        try:
            self.val = EV(C("block:block_rq_insert"), 1)
        except BadEV:
            self.val = 0
        self.thresh = self.val > 0

class NetworkTX:
    name = "NetworkTX"
    desc = " Network packets send to a network device. Aggregated (TSO/GRO) packets are counted as single packets."
    nogroup = True
    subplot = "IO"
    def compute(self, EV):
        try:
            self.val = EV(C("net:net_dev_start_xmit"), 1)
        except BadEV:
            self.val = 0
        self.thresh = self.val > 0

class NetworkRX:
    name = "NetworkRX"
    desc = " Network packets received from a network device. Aggregated (GRO) packets are counted as single packets."
    nogroup = True
    subplot = "IO"
    def compute(self, EV):
        try:
            self.val = (EV(C("net:netif_rx"), 1) +
                        EV(C("net:netif_receive_skb"), 1) +
                        EV(C("net:napi_gro_receive_entry"), 1) +
                        EV(C("net:napi_gro_frags_entry"), 1))
        except BadEV:
            self.val = 0
        self.thresh = self.val > 0

# trace events
# sched stats? (non perf)
#   yield percentage
#   % sched to idle
#   ttw local vs remote
#   wait for cpu time
# sched wakeup
# sched iowait
# filemap add to page cache
# compaction begin
# page_alloc
# start reclaim
# XXX large page alloc
# intel_gpu_freq_change
# XXX gpu
# kvm:kvm_exit

warned = False

class Setup:
    def __init__(self, r):
        r.force_metric(CS())
        r.force_metric(MinorFaults())
        r.force_metric(MajorFaults())
        r.force_metric(Migrations())
        global warned
        if os.path.exists("/sys/kernel/debug/tracing/events"):
            r.force_metric(Syscalls())
            r.force_metric(Interrupts())
            r.force_metric(IPIs())
            r.force_metric(Workqueues())
            r.force_metric(BlockIOs())
            r.force_metric(NetworkTX())
            r.force_metric(NetworkRX())
        elif sys.argv[0].find("toplev") >= 0 and "--import" not in sys.argv and not warned:
            warned = True
            print("Need to be root for trace point Linux software metrics.", file=sys.stderr)



extra_derived = {
# CBO
     "CBO.LLC_PCIE_MEM_WRITE_BYTES": {
          "Box": "CBO",
          "Category": "CBO TOR Events",
          "Defn": "LLC Miss Data from PCIe to memory written",
          "Desc": "LLC Miss Data from PCIe to memory written",
          "Equation": "TOR_INSERTS.OPCODE with:Cn_MSR_PMON_BOX_FILTER.opc=0x19e * 64",
     },
     "CBO.LLC_PCIE_MEM_READ_BYTES": {
          "Box": "CBO",
          "Category": "CBO TOR Events",
          "Defn": "LLC Miss Data from PCIe read from memory",
          "Desc": "LLC Miss Data from PCIe read from memory",
          "Equation": "TOR_INSERTS.OPCODE with:Cn_MSR_PMON_BOX_FILTER.opc=0x19c * 64",
     },
     "CBO.LLC_PCIE_MEM_TOTAL_BYTES": {
          "Box": "CBO",
          "Category": "CBO TOR Events",
          "Defn": "LLC Miss Data from PCIe read from memory",
          "Desc": "LLC Miss Data from PCIe read from memory",
          "Equation": "LLC_PCIE_MEM_READ_BYTES + LLC_PCIE_MEM_WRITE_BYTES"
     },
     "CBO.LLC_DDIO_MEM_WRITE_BYTES": {
          "Box": "CBO",
          "Category": "CBO TOR Events",
          "Defn": "LLC Miss Data from DDIO to memory written",
          "Desc": "LLC Miss Data from DDIO to memory written",
          "Equation": "TOR_INSERTS.MISS_OPCODE with:Cn_MSR_PMON_BOX_FILTER.opc=0x19e * 64",
     },
     "CBO.LLC_DDIO_MEM_READ_BYTES": {
          "Box": "CBO",
          "Category": "CBO TOR Events",
          "Defn": "LLC Miss Data from PCIe read from memory",
          "Desc": "LLC Miss Data from PCIe read from memory",
          "Equation": "TOR_INSERTS.MISS_OPCODE with:Cn_MSR_PMON_BOX_FILTER.opc=0x19c * 64",
     },
     "CBO.LLC_DDIO_MEM_TOTAL_BYTES": {
          "Box": "CBO",
          "Category": "CBO TOR Events",
          "Defn": "LLC Miss Data from DDIO read from memory",
          "Desc": "LLC Miss Data from DDIO read from memory",
          "Equation": "LLC_DDIO_MEM_READ_BYTES + LLC_DDIO_MEM_WRITE_BYTES"
     },
      "CBO.AVG_LLC_DATA_READ_MISS_LATENCY": {
          "Box": "CBO",
          "Category": "CBO CACHE Events",
          "Defn": "Average LLC data read (demand+prefetch) miss latency (core clocks)",
          "Desc": "Average LLC data read (demand+prefetch) miss latency (core clocks)",
          "Equation": "(TOR_OCCUPANCY.MISS_OPCODE / TOR_INSERTS.MISS_OPCODE) with:Cn_MSR_PMON_BOX_FILTER.opc=0x182"
     },

# PCU
    "PCU.PCT_FREQ_BAND0": {
        "Box": "PCU",
          "Category": "PCU FREQ_RESIDENCY Events",
          "Defn": "Counts the percent that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  One can use all four counters with this event, so it is possible to track up to 4 configurable bands.  One can use edge detect in conjunction with this event to track the number of times that we transitioned into a frequency greater than or equal to the configurable frequency. One can also use inversion to track cycles when we were less than the configured frequency.",
          "Desc": "Frequency Residency",
          "Notes": "The PMON control registers in the PCU only update on a frequency transition.   Changing the measuring threshold during a sample interval may introduce errors in the counts.   This is especially true when running at a constant frequency for an extended period of time.  There is a corner case here: we set this code on the GV transition.  So, if we never GV we will never call this code.  This event does not include transition times.  It is handled on fast path.",
          "Equation": "FREQ_BAND0_CYCLES / CLOCKTICKS"
     },
     "PCU.PCT_FREQ_BAND1": {
          "Box": "PCU",
          "Category": "PCU FREQ_RESIDENCY Events",
          "Defn": "Counts the percent that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  One can use all four counters with this event, so it is possible to track up to 4 configurable bands.  One can use edge detect in conjunction with this event to track the number of times that we transitioned into a frequency greater than or equal to the configurable frequency. One can also use inversion to track cycles when we were less than the configured frequency.",
          "Desc": "Frequency Residency",
          "Notes": "The PMON control registers in the PCU only update on a frequency transition.   Changing the measuring threshold during a sample interval may introduce errors in the counts.   This is especially true when running at a constant frequency for an extended period of time.  There is a corner case here: we set this code on the GV transition.  So, if we never GV we will never call this code.  This event does not include transition times.  It is handled on fast path.",
          "Equation": "FREQ_BAND1_CYCLES / CLOCKTICKS"
     },
     "PCU.PCT_FREQ_BAND2": {
          "Box": "PCU",
          "Category": "PCU FREQ_RESIDENCY Events",
          "Defn": "Counts the percent that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  One can use all four counters with this event, so it is possible to track up to 4 configurable bands.  One can use edge detect in conjunction with this event to track the number of times that we transitioned into a frequency greater than or equal to the configurable frequency. One can also use inversion to track cycles when we were less than the configured frequency.",
          "Desc": "Frequency Residency",
          "Notes": "The PMON control registers in the PCU only update on a frequency transition.   Changing the measuring threshold during a sample interval may introduce errors in the counts.   This is especially true when running at a constant frequency for an extended period of time.  There is a corner case here: we set this code on the GV transition.  So, if we never GV we will never call this code.  This event does not include transition times.  It is handled on fast path.",
          "Equation": "FREQ_BAND2_CYCLES / CLOCKTICKS"
     },
     "PCU.PCT_FREQ_BAND3": {
          "Box": "PCU",
          "Category": "PCU FREQ_RESIDENCY Events",
          "Defn": "Counts the percent that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  One can use all four counters with this event, so it is possible to track up to 4 configurable bands.  One can use edge detect in conjunction with this event to track the number of times that we transitioned into a frequency greater than or equal to the configurable frequency. One can also use inversion to track cycles when we were less than the configured frequency.",
          "Desc": "Frequency Residency",
          "Notes": "The PMON control registers in the PCU only update on a frequency transition.   Changing the measuring threshold during a sample interval may introduce errors in the counts.   This is especially true when running at a constant frequency for an extended period of time.  There is a corner case here: we set this code on the GV transition.  So, if we never GV we will never call this code.  This event does not include transition times.  It is handled on fast path.",
          "Equation": "FREQ_BAND3_CYCLES / CLOCKTICKS"
     },
     "QPI_LL.QPI_SPEED": {
          "Box": "QPI_LL",
          "Category": "QPI_LL CFCLK Events",
          "Counters": "0-3",
          "Defn": "QPI speed - GT/s",
          "Desc": "QPI speed - GT/s",
          "Equation": "CLOCKTICKS/NUM_R3QPI*8/1000000000",
     },
     "iMC.DIMM_SPEED": {
          "Box": "iMC",
          "Category": "iMC CAS Events",
          "Defn": "DIMM Speed",
          "Desc": "DIMM Speed",
          "Equation": "MC_Chy_PCI_PMON_CTR_FIXED / 2",
          "Obscure": 1,
     },


}



#!/usr/bin/env python
# print histogram for perf.data
from __future__ import print_function
import perfpd
import pfeat
import argparse

p = argparse.ArgumentParser(description='Print histogram for perf.data')
p.add_argument('datafiles', nargs='*',
               help='perf.data files (default perf.data)',
               default=['perf.data'])
p.add_argument('--sort', help='field to sort on (symbol, line)',
               default='symbol')
p.add_argument('--min-percent', help='Minimum percent to print', default=1.0)
args = p.parse_args()

COLUMN_PAD = 5
MAX_COLUMN = 70

def compute_cols(names):
    return min(max(map(len, names)) + COLUMN_PAD, MAX_COLUMN)

min_percent = float(args.min_percent) / 100.0
for d in args.datafiles:
    df, et, feat = perfpd.read_samples(d, (args.sort == 'line'))
    pfeat.print_feat(feat)

    # xxx split by event
    if 'period' in df:
        total = float(df['period'].sum())
        g = df.groupby(args.sort)
        h = g.period.sum()
        h.sort(ascending=False)
        h = h.apply(lambda x: x / total)
    else:
        h = df[args.sort].value_counts(normalize=True)
    h = h[h >= min_percent]

    cols = compute_cols(h.index)
    for s, v in zip(h.index, h.values):
        print("%-*s %.2f%%" % (cols, s, v * 100.0))



#!/usr/bin/env python3
# generate dot diagram of top down tree from module
from __future__ import print_function
import sys

max_level = 5
first = 1
if sys.argv[1:] and sys.argv[1][:2] == "-l":
    max_level = int(sys.argv[1][2:])
    first += 1
if len(sys.argv) > first and sys.argv[first] == "simple":
    import simple_ratios
    m = simple_ratios
else:
    import adl_glc_ratios
    m = adl_glc_ratios

def has(obj, name):
    return name in obj.__class__.__dict__

class Runner:
    def __init__(self):
        self.olist = []

    def run(self, n):
        if n.level <= max_level:
            self.olist.append(n)

    def metric(self, n):
        pass

    def finish(self):
        for n in self.olist:
            if n.level == 1:
                print('"%s";' % (n.name))
            elif n.parent:
                print('"%s" -> "%s";' % (n.parent.name, n.name))
            #if n.sibling:
            #    print('"%s" -> "%s";' % (n.name, n.sibling.name))

runner = Runner()
m.Setup(runner)
print(runner.olist, file=sys.stderr)
print("digraph {")
print("fontname=\"Courier\";")
runner.finish()
print("}")



#!/usr/bin/env python3
# library and tool to access Intel MSRs (model specific registers)
# Author: Andi Kleen
from __future__ import print_function
import glob
import struct
import os

def writemsr(msr, val):
    n = glob.glob('/dev/cpu/[0-9]*/msr')
    for c in n:
        f = os.open(c, os.O_WRONLY)
        os.lseek(f, msr, os.SEEK_SET)
        os.write(f, struct.pack('Q', val))
        os.close(f)
    if not n:
        raise OSError("msr module not loaded (run modprobe msr)")

def readmsr(msr, cpu = 0):
    f = os.open('/dev/cpu/%d/msr' % (cpu,), os.O_RDONLY)
    os.lseek(f, msr, os.SEEK_SET)
    val = struct.unpack('Q', os.read(f, 8))[0]
    os.close(f)
    return val

def changebit(msr, bit, val):
    n = glob.glob('/dev/cpu/[0-9]*/msr')
    for c in n:
        f = os.open(c, os.O_RDWR)
        os.lseek(f, msr, os.SEEK_SET)
        v = struct.unpack('Q', os.read(f, 8))[0]
        if val:
            v = v | (1 << bit)
        else:
            v = v & ~(1 << bit)
        os.lseek(f, msr, os.SEEK_SET)
        os.write(f, struct.pack('Q', v))
        os.close(f)
    if not n:
        raise OSError("msr module not loaded (run modprobe msr)")

if __name__ == '__main__':
    import argparse

    def parse_hex(s):
        try:
            return int(s, 16)
        except ValueError:
            raise argparse.ArgumentError("Bad hex number %s" % (s))

    if not os.path.exists("/dev/cpu/0/msr"):
        os.system("/sbin/modprobe msr")

    p = argparse.ArgumentParser(description='Access x86 model specific registers.')
    p.add_argument('msr', type=parse_hex, help='number of the MSR to access')
    p.add_argument('value', nargs='?', type=parse_hex, help='value to write (if not specified read)')
    p.add_argument('--setbit', type=int, help='Bit number to set')
    p.add_argument('--clearbit', type=int, help='Bit number to clear')
    p.add_argument('--cpu', type=int, default=0, help='CPU to read on (writes always change all)')
    args = p.parse_args()
    if args.value is None and not args.setbit and not args.clearbit:
        print("%x" % (readmsr(args.msr, args.cpu)))
    elif args.setbit:
        changebit(args.msr, args.setbit, 1)
    elif args.clearbit:
        changebit(args.msr, args.clearbit, 0)
    else:
        writemsr(args.msr, args.value)



# generic utilities for lists
import sys
from itertools import chain

if sys.version_info.major == 3:
    from itertools import zip_longest
else:
    from itertools import izip_longest
    zip_longest = izip_longest

def flatten(x):
    return list(chain(*x))

def filternot(p, l):
    return list(filter(lambda x: not p(x), l))

# add items from b to a if not already in a
def cat_unique(a, b):
    aset = set(a)
    add = [x for x in b if x not in aset]
    return a + add

# remove duplicates without reordering
def dedup(a):
    l = []
    prev = set()
    for j in a:
        if j not in prev:
            l.append(j)
            prev.add(j)
    return l

def not_list(l):
    return [not x for x in l]

# merge two dicts with appending lists
def append_dict(a, b):
    for k in b:
        if k in a:
            a[k] += b[k]
        else:
            a[k] = b[k]

# create dict/list with same shape as a, but filled with dummy values
def dummy_dict(a, val=0.0):
    return {k: [val] * len(a[k]) for k in a}

def padlist(l, length, val=0.0):
    if len(l) < length:
        return l + [val]*(length-len(l))
    return l

def findprefix(l, prefix, stop=None):
    for i, v in enumerate(l):
        if v == stop:
            break
        if v.startswith(prefix):
            return i
    return -1
